from functools import partial
from pathlib import Path
from typing import Dict, Optional, Union, cast

import numpy as np
import pysam
import seqpro as sp
from numpy.typing import ArrayLike, NDArray
from typing_extensions import assert_never

from .types import Reader
from .util import get_rel_starts, normalize_contig_name


class NoPadError(Exception):
    pass


class Fasta(Reader):
    dtype = np.dtype("S1")
    sizes = {}
    coords = {}
    chunked = False

    def __init__(
        self,
        name: str,
        path: Union[str, Path],
        pad: Optional[str] = None,
        alphabet: Optional[Union[str, sp.NucleotideAlphabet, sp.AminoAlphabet]] = None,
        in_memory=False,
    ) -> None:
        """Read sequences from a FASTA file.

        Parameters
        ----------
        name : str
            Name of the reader, for example `'seq'`.
        path : Union[str, Path]
            Path to the FASTA file.
        pad : Optional[str], optional
            A single character which, if passed, will pad out-of-bound ranges with this
            value. By default no padding is done and out-of-bound ranges raise an error.
        alphabet : str, sp.NucleotideAlphabet, sp.AminoAlphabet, optional
            Alphabet to use for the sequences. If not passed, defaults to DNA.

        Raises
        ------
        ValueError
            If pad value is not a single character.
        """
        self.name = name
        self.path = path
        if pad is None:
            self.pad = pad
        else:
            if len(pad) > 1:
                raise ValueError("Pad value must be a single character.")
            self.pad = pad.encode("ascii")

        with self._open() as f:
            self.contigs = {c: f.get_reference_length(c) for c in f.references}

        if alphabet is None:
            self.alphabet: Union[
                sp.NucleotideAlphabet, sp.AminoAlphabet
            ] = sp.alphabets.DNA
        elif isinstance(alphabet, str):
            alphabet = alphabet.upper()
            try:
                self.alphabet = cast(
                    Union[sp.NucleotideAlphabet, sp.AminoAlphabet],
                    getattr(sp.alphabets, alphabet),
                )
            except AttributeError:
                raise ValueError(f"Alphabet {alphabet} not found.")
        else:
            self.alphabet = alphabet

        if isinstance(self.alphabet, sp.NucleotideAlphabet):
            self.rev_strand_fn = partial(
                self.alphabet.reverse_complement, length_axis=-1
            )
        elif isinstance(self.alphabet, sp.AminoAlphabet):

            def rev_strand_fn(a: NDArray[np.bytes_]):
                return a[::-1]

            self.rev_strand_fn = rev_strand_fn  # type: ignore[assignment]
        else:
            assert_never(self.alphabet)

        self.contigs = self._get_contig_lengths()

        self.handle: Optional[pysam.FastaFile] = None

        if not in_memory:
            self.sequences = None
        else:
            self.sequences = self._get_all_contigs()

    def _get_contig_lengths(self) -> Dict[str, int]:
        with self._open() as f:
            return {c: f.get_reference_length(c) for c in f.references}

    def _get_all_contigs(self) -> Dict[str, NDArray[np.bytes_]]:
        """Load all contigs into memory."""
        with self._open() as f:
            return {
                c: np.frombuffer(f.fetch(c).encode("ascii"), "S1") for c in f.references
            }

    def _open(self):
        return pysam.FastaFile(str(self.path))

    def close(self):
        if self.handle is not None:
            self.handle.close()
        self.handle = None

    def read(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        **kwargs,
    ) -> NDArray[np.bytes_]:
        """Read a sequence from a FASTA file.

        Parameters
        ----------
        contig : str
            Name of the contig/chromosome.
        starts : ArrayLike
            Start coordinates, 0-based.
        ends : ArrayLike
            End coordinates, 0-based exclusive.
        **kwargs
            Not used.

        Returns
        -------
        NDArray[np.bytes_]
            Sequence from FASTA file.

        Raises
        ------
        ValueError
            Coordinates are out-of-bounds and pad value is not set.
        """
        _contig = normalize_contig_name(contig, self.contigs)
        if _contig is None:
            raise ValueError(f"Contig {contig} not found.")
        else:
            contig = _contig
        contig_len = self.contigs[contig]

        starts = np.atleast_1d(np.asarray(starts, dtype=np.int64))
        ends = np.atleast_1d(np.asarray(ends, dtype=np.int64))

        if starts.ndim > 1 or ends.ndim > 1:
            raise ValueError("Starts and ends must be coercible to 1D arrays.")

        if len(starts) != len(ends):
            raise ValueError("Starts and ends must be the same length.")

        lengths = ends - starts

        q_starts = np.maximum(0, starts)
        q_ends = np.minimum(self.contigs[contig], ends)

        left_pads = -np.minimum(0, starts)
        right_pads = np.maximum(0, ends - contig_len)

        if self.pad is None:
            if (left_pads > 0).any():
                raise NoPadError("Padding is disabled and a start is < 0.")
            if (right_pads > 0).any():
                raise NoPadError("Padding is disabled and an end is > contig length.")
            out = np.empty(lengths.sum(), dtype="S1")
        else:
            out = np.full(lengths.sum(), self.pad, dtype="S1")

        rel_starts = get_rel_starts(starts, ends)
        rel_ends = rel_starts + lengths - right_pads
        rel_starts += left_pads

        if self.sequences is None:
            if self.handle is None:
                self.handle = self._open()

            for q_start, q_end, rel_start, rel_end in zip(
                q_starts, q_ends, rel_starts, rel_ends
            ):
                seq = self.handle.fetch(contig, q_start, q_end)
                seq = np.frombuffer(seq.encode("ascii"), "S1")
                out[rel_start:rel_end] = seq
        else:
            for q_start, q_end, rel_start, rel_end in zip(
                q_starts, q_ends, rel_starts, rel_ends
            ):
                seq = self.sequences[contig][q_start:q_end]
                out[rel_start:rel_end] = seq

        return out

    def vidx(
        self, contigs: ArrayLike, starts: ArrayLike, length: int, **kwargs
    ) -> NDArray[np.bytes_]:
        # TODO: check that all queries are within bounds if pad is None

        unique_contigs, c_idx = np.unique(contigs, return_inverse=True)
        renamed_contigs = np.array(
            [normalize_contig_name(c, self.contigs) for c in unique_contigs]
        )
        if (renamed_contigs == None).any():  # noqa: E711
            raise RuntimeError("Contig not found in FASTA file.")
        contigs = renamed_contigs[c_idx]

        starts = np.atleast_1d(np.asarray(starts, dtype=np.int64))
        ends = starts + length

        if self.pad is None:
            out = np.empty((len(starts), length), dtype="S1")
        else:
            out = np.full((len(starts), length), self.pad, dtype="S1")

        for i, (c, s, e) in enumerate(zip(contigs, starts, ends)):
            if self.sequences is None:
                if self.handle is None:
                    self.handle = self._open()
                seq = self.handle.fetch(c, s, e)
                seq = np.frombuffer(seq.encode("ascii"), "S1")
            else:
                seq = self.sequences[c][s:e]
            out[i] = seq

        return out



from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import numpy as np
from attrs import define
from numpy.typing import ArrayLike, NDArray

from ..util import normalize_contig_name
from .genotypes import (
    Genotypes,
    MemmapGenos,
    NumpyGenos,
    PgenGenos,
    VCFGenos,
    VIdxGenos,
    ZarrGenos,
)
from .records import Records, VLenAlleles

__all__ = [
    "PgenGenos",
    "ZarrGenos",
    "MemmapGenos",
    "VCFGenos",
    "Variants",
    "Records",
]


@define
class DenseGenotypes:
    """Dense array(s) of genotypes.

    Attributes
    ----------
    positions : NDArray[np.int32]
        Shape: (variants)
    size_diffs : NDArray[np.int32]
        Shape : (variants). Difference in length between the REF and the ALT alleles.
    ref : VLenAlleles
        Shape: (variants). REF alleles.
    alt : VLenAlleles
        Shape: (variants). ALT alleles.
    genotypes : NDArray[np.int8]
        Shape: (samples, ploid, variants)
    offsets : NDArray[np.uint32], optional
        Shape: (regions + 1). Offsets for the index boundaries of each region such
        that variants for region `i` are `positions[offsets[i] : offsets[i+1]]`,
        `size_diffs[offsets[i] : offsets[i+1]]`, ..., etc.
    """

    positions: NDArray[np.int32]
    size_diffs: NDArray[np.int32]
    ref: VLenAlleles
    alt: VLenAlleles
    genotypes: NDArray[np.int8]
    offsets: NDArray[np.uint32]


@define
class Variants:
    records: Records
    genotypes: Genotypes
    _sample_idxs: Optional[NDArray[np.intp]] = None

    @property
    def chunked(self):
        return self.genotypes.chunked

    @property
    def samples(self):
        if self._sample_idxs is None:
            return self.genotypes.samples
        return self.genotypes.samples[self._sample_idxs]

    @property
    def n_samples(self):
        if self._sample_idxs is None:
            return self.genotypes.n_samples
        return len(self._sample_idxs)

    @property
    def ploidy(self):
        return self.genotypes.ploidy

    @classmethod
    def from_vcf(
        cls,
        vcf: Union[str, Path, Dict[str, Path]],
        use_cache: bool = True,
        chunk_shape: Optional[Tuple[int, int, int]] = None,
    ):
        """Currently does not support multi-allelic sites, but does support *split*
        multi-allelic sites. Note that SVs and "other" variants are also not supported.
        VCFs can be prepared by running:
        ```bash
        bcftools view -i 'TYPE="snp" || TYPE="indel"' <file.bcf> \\
        | bcftools norm \\
            -a \\
            --atom-overlaps . \\
            -m - \\
            -f <ref.fa> \\
            -O b \\
            -o <norm.bcf>
        ```
        """
        records = Records.from_vcf(vcf)

        if use_cache:
            try:
                genotypes = ZarrGenos(vcf)
            except FileNotFoundError:
                genotypes = VCFGenos(vcf, records.contig_offsets)
                genotypes = ZarrGenos.from_recs_genos(
                    records, genotypes, chunk_shape=chunk_shape
                )
        else:
            genotypes = VCFGenos(vcf, records.contig_offsets)
        return cls(records, genotypes)

    @classmethod
    def from_gvl(cls, path: Union[str, Path, Dict[str, Path]]):
        """Construct a Variants object from GVL files. The path(s) must end with `.gvl`.

        Parameters
        ----------
        path : Union[str, Path, Dict[str, Path]]
            Path to the GVL file(s).

        Returns
        -------
        Variants
        """
        records = Records.from_gvl_arrow(path)
        genotypes = ZarrGenos(path)
        return cls(records, genotypes)

    # TODO: read sample names from .psam file by using #IID or IID column. Implement a .psam reader.
    @classmethod
    def from_pgen(
        cls, pgen: Union[str, Path, Dict[str, Path]], sample_names: ArrayLike
    ):
        """Currently does not support multi-allelic sites, but does support *split*
        multi-allelic sites. Note that SVs and "other" variants are also not supported.
        A PGEN can be prepared from a VCF by running:
        ```bash
        bcftools view -i 'TYPE="snp" || TYPE="indel"' <file.bcf> \\
        | bcftools norm \\
            -a \\
            --atom-overlaps . \\
            -m - \\
            -f <ref.fa> \\
            -O b \\
            -o <norm.bcf> \\
        plink2 --make-pgen \\
            --bcf <norm.bcf> \\
            --vcf-half-call r \\
            --out <prefix>
        ```
        """
        if isinstance(pgen, str):
            pgen = Path(pgen)

        if isinstance(pgen, Path):
            pgen = {"_all": pgen}
        records = Records.from_pvar(
            {c: p.with_suffix(".pvar") for c, p in pgen.items()}
        )
        sample_names = np.atleast_1d(np.asarray(sample_names))
        genotypes = PgenGenos(pgen, sample_names)
        return cls(records, genotypes)

    def in_memory(self):
        if not isinstance(self.genotypes, NumpyGenos):
            genotypes = NumpyGenos.from_recs_genos(self.records, self.genotypes)
            return self.__class__(self.records, genotypes)
        return self

    def subset_samples(self, samples: ArrayLike):
        samples = np.atleast_1d(np.asarray(samples, dtype=str))
        geno_sample_idxs, sample_idxs = np.intersect1d(
            self.genotypes.samples, samples, return_indices=True
        )[1:]
        if len(sample_idxs) != len(samples):
            raise ValueError("Some samples were not found")
        self._sample_idxs = geno_sample_idxs[sample_idxs]

    def read(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        sample: Optional[ArrayLike] = None,
        ploid: Optional[ArrayLike] = None,
    ):
        contig = normalize_contig_name(
            contig, self.records.contigs
        )  # pyright: ignore[reportAssignmentType]
        if contig is None:
            return None

        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = np.atleast_1d(np.asarray(ends, dtype=int))

        recs = self.records.vars_in_range(contig, starts, ends)
        if recs is None:
            return None

        if sample is not None:
            sample = np.atleast_1d(np.asarray(sample, dtype=str))
            geno_sample_idxs, sample_idxs = np.intersect1d(
                self.genotypes.samples, sample, return_indices=True
            )[1:]
            sample_idxs = geno_sample_idxs[sample_idxs]
            if len(sample_idxs) != len(sample):
                raise ValueError("Some samples were not found")
        else:
            sample_idxs = None

        if ploid is not None:
            ploid = np.atleast_1d(np.asarray(ploid, dtype=int))

        # (s p v)
        genos = self.genotypes.read(
            contig, recs.start_idxs, recs.end_idxs, sample_idxs, ploid
        )

        return DenseGenotypes(
            recs.positions,
            recs.size_diffs,
            recs.refs,
            recs.alts,
            genos,
            recs.offsets,
        )

    def read_for_haplotype_construction(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        sample: Optional[ArrayLike] = None,
        ploid: Optional[ArrayLike] = None,
    ) -> Tuple[Optional[DenseGenotypes], NDArray[np.int32]]:
        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = np.atleast_1d(np.asarray(ends, dtype=int))

        _contig = normalize_contig_name(
            contig, self.records.contigs
        )  # pyright: ignore[reportAssignmentType]
        if _contig is None:
            return None, ends.astype(np.int32)
        else:
            contig = _contig

        recs, max_ends = self.records.vars_in_range_for_haplotype_construction(
            contig, starts, ends
        )
        if recs is None:
            return None, ends

        if sample is not None:
            sample = np.atleast_1d(np.asarray(sample, dtype=str))
            geno_sample_idxs, sample_idxs = np.intersect1d(
                self.genotypes.samples, sample, return_indices=True
            )[1:]
            sample_idxs = geno_sample_idxs[sample_idxs]
            if len(sample_idxs) != len(sample):
                raise ValueError("Some samples were not found")
        else:
            sample_idxs = None

        if ploid is not None:
            ploid = np.atleast_1d(np.asarray(ploid, dtype=int))

        # (s p v)
        genos = self.genotypes.read(
            contig, recs.start_idxs, recs.end_idxs, sample_idxs, ploid
        )

        return (
            DenseGenotypes(
                recs.positions,
                recs.size_diffs,
                recs.refs,
                recs.alts,
                genos,
                recs.offsets,
            ),
            max_ends,
        )

    def vidx(
        self,
        contigs: ArrayLike,
        starts: ArrayLike,
        length: int,
        samples: ArrayLike,
        ploidies: ArrayLike,
    ):
        if not isinstance(self.genotypes, VIdxGenos):
            raise ValueError(
                f"Genotypes {self.genotypes} does not support vectorized indexing."
            )

        contigs = np.atleast_1d(np.asarray(contigs, dtype=str))
        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        samples = np.atleast_1d(np.asarray(samples, dtype=str))
        ploidies = np.atleast_1d(np.asarray(ploidies, dtype=int))

        recs = self.records.vidx_vars_in_range(contigs, starts, length)
        if recs is None:
            return None

        genos = self.genotypes.vidx(
            contigs, recs.start_idxs, recs.end_idxs, samples, ploidies
        )

        return DenseGenotypes(
            recs.positions,
            recs.size_diffs,
            recs.refs,
            recs.alts,
            genos,
            recs.offsets,
        )

    def vidx_for_haplotype_construction(
        self,
        contigs: ArrayLike,
        starts: ArrayLike,
        length: int,
        samples: ArrayLike,
        ploidies: ArrayLike,
    ):
        if not isinstance(self.genotypes, VIdxGenos):
            raise ValueError(
                f"Genotypes {self.genotypes} does not support vectorized indexing."
            )

        contigs = np.atleast_1d(np.asarray(contigs, dtype=np.str_))
        starts = np.atleast_1d(np.asarray(starts, dtype=np.int32))
        samples = np.atleast_1d(np.asarray(samples, dtype=np.str_))
        ploidies = np.atleast_1d(np.asarray(ploidies, dtype=np.intp))

        recs, max_ends = self.records.vidx_vars_in_range_for_haplotype_construction(
            contigs, starts, length
        )
        if recs is None:
            return None, max_ends

        unique_samples, inverse = np.unique(samples, return_inverse=True)
        if missing := set(unique_samples).difference(self.samples):
            raise ValueError(f"Samples {missing} were not found")
        key_idx, query_idx = np.intersect1d(
            self.samples, unique_samples, return_indices=True, assume_unique=True
        )[1:]
        sample_idx = key_idx[query_idx[inverse]]

        if (ploidies >= self.ploidy).any():
            raise ValueError("Ploidies requested exceed maximum ploidy")

        genos = self.genotypes.vidx(
            contigs, recs.start_idxs, recs.end_idxs, sample_idx, ploidies
        )

        return (
            DenseGenotypes(
                recs.positions,
                recs.size_diffs,
                recs.refs,
                recs.alts,
                genos,
                recs.offsets,
            ),
            max_ends,
        )



from pathlib import Path
from typing import Any, Dict, Mapping, Optional, Sequence, Union

import numba as nb
import numpy as np
import polars as pl
import xarray as xr
from numpy.typing import NDArray

from .types import Reader


class Intervals(Reader):
    chunked = False

    def __init__(
        self,
        name: str,
        table: Union[str, Path, pl.DataFrame],
        subset: Optional[Mapping[str, Sequence[Any]]] = None,
        value_columns: Optional[Sequence[str]] = None,
        value_dim: Optional[str] = None,
        validate=True,
        zero_indexed=True,
        lazy=True,
        **table_reader_kwargs,
    ):
        """Read values from a table of intervals. This corresponds to the
        UCSC BED3+ format, the first three columns must be `chrom`, `chromStart`, and
        `chromEnd`, and other columns can be anything. By default, the 4th column is
        used as the value column, but this can be changed with `value_columns`. Any
        columns that aren't a value column are treated as grouping columns, and can be
        used to subset the table with `subset`. The order of the columns will determine
        the order of the dimensions in the output array:
        - If there are no grouping columns, the output array will have shape
        `(n_values, length)` and the order will match the order of `value_columns`.
        - If there are grouping columns, the output array will have shape
        `(group_1, ..., group_n, n_values, length)` and the order of each group will
        match the order of `subset`.

        Parameters
        ----------
        name : str
            Name of the reader.
        table : Union[str, Path, pl.DataFrame]
            DataFrame, or path to one, in UCSC BED3+ format.
        subset : Mapping[str, Sequence[str]], optional
            Mapping of column names to groups of interest, by default None (all groups
            used). For example, if there was a column "name" with sample names, to
            only include samples "A" and "B" `subset={"name": ["A", "B"]}`.
        value_columns : Optional[Sequence[str]], optional
            Which columns to use as values, by default None (4th column used).
        validate : bool, optional
            Whether to validate the RLE table, by default True.
        zero_indexed : bool, optional
            Whether the coordinates are 0-indexed, by default True
        lazy : bool, optional
            Whether to hold the table in-memory or query it lazily, leaving it on-disk
            to reduce memory usage at the cost of query speed. Ignored if `table` is a
            DataFrame.
        **table_reader_kwargs
            Passed to the table reader function (either pl.scan_csv or pl.scan_ipc). This
            can be used to rename columns.
        """

        if isinstance(table, (str, Path)):
            table = Path(table)
            suffixes = set(table.suffixes)
            reader_kwargs: Dict[str, Any] = {}
            reader_kwargs.update(table_reader_kwargs)
            if ".csv" in suffixes:
                reader_kwargs["separator"] = ","
                reader = pl.scan_csv
            elif {".txt", ".tsv"} & suffixes:
                reader_kwargs["separator"] = "\t"
                reader = pl.scan_csv
            elif {".fth", ".feather", ".ipc", ".arrow"} & suffixes:
                reader = pl.scan_ipc
            else:
                raise ValueError(f"Table has unrecognized file extension: {table.name}")
            _table = reader(table, **reader_kwargs)
        else:
            _table = table.lazy()

        if not lazy:
            _table = _table.collect()

        if not zero_indexed:
            _table = _table.with_columns(pl.col("chromStart") - 1)

        if value_columns is None:
            value_columns = [_table.columns[3]]

        if subset is None:
            grouping_cols = get_grouping_cols(_table, value_columns)
            _subset = {
                c: _table.lazy()
                .select(c)
                .unique(maintain_order=True)
                .collect()[c]
                .to_numpy()
                for c in grouping_cols
            }
        else:
            grouping_cols = [c for c in subset.keys()]
            for col, groups in subset.items():
                _table = _table.filter(pl.col(col).is_in(groups))
            _subset = {c: np.asarray(group) for c, group in subset.items()}

        if value_dim is None:
            value_dim = "values"

        _subset[value_dim] = np.asarray(value_columns)

        if validate:
            _table = validate(_table, grouping_cols, value_columns).lazy()

        self.name = name
        self.value_dim = value_dim
        self.value_columns = value_columns
        self.table = _table
        self.dtype = (
            _table.lazy().select(value_columns[0]).head(1).collect().to_numpy().dtype
        )
        self.coords = _subset
        self.sizes = {c: len(g) for c, g in _subset.items()}
        contigs = self.table.lazy().select(pl.col("chrom").unique()).collect()["chrom"]
        self.contig_starts_with_chr = self.infer_contig_prefix(contigs)

    def rev_strand_fn(self, arr: NDArray):
        return arr[..., ::-1]

    def read(self, contig: str, start: int, end: int, **kwargs) -> xr.DataArray:
        groups = {}
        sizes = {}
        for c in self.coords:
            if c == self.value_dim:
                continue
            groups[c] = kwargs.get(c, self.coords[c])
            sizes[c] = len(groups[c])
        subset = bool(set(kwargs.keys()) & set(self.coords.keys()))

        contig = self.normalize_contig_name(contig)
        length = end - start

        q = self.table.lazy().filter(
            (pl.col("chrom") == contig)
            & (pl.col("chromStart") < end)
            & (pl.col("chromEnd") > start)
        )

        # get groups in order requested
        if subset:
            q = q.filter(*(pl.col(c).is_in(g) for c, g in groups.items()))
            with pl.StringCache():
                for c, g in groups.items():
                    pl.Series(g, dtype=pl.Categorical)
                q = q.sort(
                    pl.col(*groups.keys()).cast(pl.Categorical),
                    "chromStart",
                    "chromEnd",
                )

        q = q.collect()

        cols = q.select(
            # TODO: generalize getting group idx to multiple columns
            pl.col("name").rle_id().alias("sample_idx"),
            pl.col("chromStart") - start,
            pl.col("chromEnd") - start,
            *self.value_columns,
        ).get_columns()

        out = np.zeros(
            shape=(*self.sizes.values(), len(self.value_columns), length),
            dtype=self.dtype,
        ).squeeze()
        sample_idx, starts, ends, vals = [c.to_numpy() for c in cols]
        assign_vals(out, sample_idx, starts, ends, vals)

        dims = list(self.sizes.keys()) + ["values", "length"]
        return xr.DataArray(out, dims=dims)


def get_grouping_cols(
    df: Union[pl.DataFrame, pl.LazyFrame], value_columns: Sequence[str]
):
    bed3_cols = ["chrom", "chromStart", "chromEnd"]
    grouping_cols = [c for c in df.columns if c not in list(value_columns) + bed3_cols]
    return grouping_cols


def validate(
    df: Union[pl.DataFrame, pl.LazyFrame],
    grouping_cols: Sequence[str],
    value_columns: Sequence[str],
):
    _df = df.lazy()

    df_bed3_schema = _df.select("chrom", "chromStart", "chromEnd").schema

    bed3_schema = {
        "chrom": pl.Utf8,
        "chromStart": pl.Int64,
        "chromEnd": pl.Int64,
    }

    missing_cols = set(bed3_schema.keys()) - set(df_bed3_schema.keys())
    if len(missing_cols) > 0:
        raise ValueError(
            f"""RLE table is missing (or has misnamed) BED3 columns:
            {missing_cols}"""
        )

    mismatched_dtypes = {
        c: (dt1, df_bed3_schema[c])
        for c, dt1 in bed3_schema.items()
        if dt1 != df_bed3_schema[c]
    }
    if mismatched_dtypes:
        raise ValueError(
            f"""RLE table has incorrect dtypes for BED3 columns.
            Mismatched dtypes (expected, seen): {mismatched_dtypes}"""
        )

    val_schema = _df.select(*value_columns).schema
    first_dtype = next(iter(val_schema.values()))
    mismatched_dtypes = {
        c: (first_dtype, dt1) for c, dt1 in val_schema.items() if dt1 != first_dtype
    }
    if mismatched_dtypes:
        raise ValueError(
            f"""RLE table has incorrect dtypes for value columns.
            Mismatched dtypes (expected, seen): {mismatched_dtypes}"""
        )

    non_numeric_cols = [c for c, dt in val_schema.items() if not dt.is_numeric()]
    if non_numeric_cols:
        raise ValueError(
            f"""RLE table has non-numeric value columns.
            Value columns: {non_numeric_cols}"""
        )

    _df = _df.collect()
    if _df.height == 0:
        raise ValueError("No entries in table.")

    null_count = _df.null_count()
    if null_count.to_numpy().sum() > 0:
        raise ValueError(
            f"Table has null values. Number of nulls per column:\n{null_count}"
        )

    duplicated = _df.is_duplicated()
    if duplicated.any():
        raise ValueError(f"Table has {duplicated.sum()} duplicate entries.")

    overlaps = (
        _df.lazy()
        .sort("chromStart", "chromEnd")
        .groupby(*grouping_cols, "chrom")
        .agg(
            pl.col("chromStart", "chromEnd"),
            (
                pl.col("chromStart").shift(-1, fill_value=np.inf) < pl.col("chromEnd")
            ).alias("overlapping"),
        )
        .explode("chromStart", "chromEnd", "overlapping")
        .filter(pl.col("overlapping") | pl.col("overlapping").shift(1))
        .collect()
    )
    if overlaps.height > 0:
        raise ValueError(f"Found overlapping intervals in the table:\n{overlaps}")

    return _df


@nb.njit(nogil=True, parallel=True, cache=True)
def assign_vals(out, sample_idx, starts, ends, vals):
    # all intervals in an RLE table are guaranteed to be non-overlapping
    for i in nb.prange(len(starts)):
        out[sample_idx[i], starts[i] : ends[i]] = vals[i]



import re
from pathlib import Path
from typing import Dict, List, Optional, Protocol, Tuple, Union, cast, runtime_checkable

import numpy as np
from numpy.typing import ArrayLike, NDArray
from tqdm.auto import tqdm
from typing_extensions import Self

from genvarloader.util import get_rel_starts

from .records import Records

try:
    import pgenlib

    PGENLIB_INSTALLED = True
except ImportError:
    PGENLIB_INSTALLED = False

try:
    import tensorstore as ts
    import zarr

    ZARR_TENSORSTORE_INSTALLED = True
except ImportError:
    ZARR_TENSORSTORE_INSTALLED = False

try:
    import cyvcf2

    CYVCF2_INSTALLED = True
except ImportError:
    CYVCF2_INSTALLED = False


class Genotypes(Protocol):
    chunked: bool
    samples: NDArray[np.str_]
    ploidy: int

    def read(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
    ) -> NDArray[np.int8]:
        """Read genotypes from a contig from index i to j, 0-based exclusive.

        Parameters
        ----------
        contig : str
            Name of the contig/chromosome.
        start_idxs : NDArray[np.intp]
            Start indices, 0-based.
        end_idxs : NDArray[np.intp]
            End indices, 0-based exclusive.
        sample_idx : NDArray[np.intp], optional
            Indices of the samples to include. Must be unique.
        haplotype_idx : NDArray[np.intp], optional
            Indices of the haplotypes to include. Must be unique.

        Returns
        -------
        genotypes : NDArray[np.int8]
            Shape: (samples ploidy variants). Genotypes for each query region.
        """
        ...

    @property
    def n_samples(self) -> int:
        return len(self.samples)


class FromRecsGenos(Protocol):
    @classmethod
    def from_recs_genos(
        cls,
        records: Records,
        genotypes: Genotypes,
        paths: Optional[Dict[str, Path]] = None,
        mem: int = int(1e9),
        overwrite: bool = False,
        **kwargs,
    ) -> Self: ...


@runtime_checkable
class VIdxGenos(Protocol):
    """Vectorized indexing for genotypes."""

    def vidx(
        self,
        contigs: ArrayLike,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idxs: ArrayLike,
        haplotype_idxs: ArrayLike,
    ) -> NDArray[np.int8]:
        """Read data akin to vectorized indexing. Output shape will be (n_variants,)."""
        ...


class PgenGenos(Genotypes):
    chunked = False
    ploidy = 2

    def __init__(
        self, paths: Union[str, Path, Dict[str, Path]], sample_names: NDArray[np.str_]
    ) -> None:
        if not PGENLIB_INSTALLED:
            raise ImportError(
                "pgenlib must be installed to use PGEN files for genotypes."
            )
        if isinstance(paths, (str, Path)):
            paths = {"_all": Path(paths)}

        if len(paths) == 0:
            raise ValueError("No paths provided.")
        n_samples = None
        n_variants = 0
        for p in paths.values():
            if not p.exists():
                raise FileNotFoundError(f"PGEN file {p} not found.")
            if n_samples is None:
                n_samples = pgenlib.PgenReader(bytes(p)).get_raw_sample_ct()  # pyright: ignore
            n_variants += pgenlib.PgenReader(bytes(p)).get_variant_ct()  # pyright: ignore
        self.paths = paths
        self.n_samples: int = n_samples  # type: ignore
        self.n_variants = n_variants
        self.samples = sample_names
        self.handle = None

    def _pgen(self, contig: str, sample_idx: Optional[NDArray[np.uint32]]):
        return pgenlib.PgenReader(bytes(self.paths[contig]), sample_subset=sample_idx)  # pyright: ignore

    def read(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
    ) -> NDArray[np.int8]:
        if self.handle is None:
            self.handle = self._pgen(contig, None)

        start_idxs = np.atleast_1d(np.asarray(start_idxs, dtype=np.int32))
        end_idxs = np.atleast_1d(np.asarray(end_idxs, dtype=np.int32))

        n_vars = (end_idxs - start_idxs).sum()
        # (v s*2)
        genotypes = np.empty((n_vars, self.n_samples * self.ploidy), dtype=np.int32)

        for i, (s, e) in enumerate(zip(start_idxs, end_idxs)):
            if s == e:
                continue
            rel_s = s - start_idxs[i]
            rel_e = e - start_idxs[i]
            self.handle.read_alleles_range(
                s, e, allele_int32_out=genotypes[rel_s:rel_e]
            )

        # (v s*2)
        genotypes = genotypes.astype(np.int8)
        # (s*2 v)
        genotypes = genotypes.swapaxes(0, 1)
        # (s 2 v)
        genotypes = np.stack([genotypes[::2], genotypes[1::2]], axis=1)

        if sample_idx is not None:
            _sample_idx = np.atleast_1d(np.asarray(sample_idx, dtype=np.intp))
        else:
            _sample_idx = slice(None)

        if haplotype_idx is not None:
            _haplotype_idx = np.atleast_1d(np.asarray(haplotype_idx, dtype=np.intp))
        else:
            _haplotype_idx = slice(None)

        genotypes = genotypes[_sample_idx, _haplotype_idx]
        # re-order samples to be in query order
        # if sample_sorter is not None and (np.arange(n_samples) != sample_sorter).any():
        #     genotypes = genotypes[sample_sorter]

        return genotypes

    def __del__(self) -> None:
        if self.handle is not None:
            self.handle.close()


"""
Implementation note regarding using tensorstore with num_workers > 0
https://github.com/google/tensorstore/issues/61
TL;DR TensorStore is not Send.
"""


class ZarrGenos(Genotypes, FromRecsGenos, VIdxGenos):
    chunked = True
    driver = "zarr"

    def __init__(self, paths: Union[str, Path, Dict[str, Path]]) -> None:
        if not ZARR_TENSORSTORE_INSTALLED:
            raise ImportError(
                "Zarr and TensorStore must be installed to use chunked array caches like Zarr and N5."
            )
        if isinstance(paths, (str, Path)):
            paths = {"_all": Path(paths)}

        if not all(p.suffix == ".zarr" for p in paths.values()):
            paths = self.convert_paths_to_zarr(paths)
            if not all(p.exists() for p in paths.values()):
                raise FileNotFoundError(f"Zarr file(s) not found: {paths}")

        self.paths = paths

        first_path = next(iter(self.paths.values()))
        z = zarr.open_array(first_path)  # pyright: ignore[reportPossiblyUnboundVariable]
        self.samples = np.asarray(z.attrs["samples"])
        # (s p v)
        self.ploidy = z.shape[1]
        self.tstores = None

    @staticmethod
    def convert_paths_to_zarr(paths: Dict[str, Path]) -> Dict[str, Path]:
        extension = re.compile(r"\.\w+(\.gz)?$")
        paths = {
            c: p.parent / extension.sub(".gvl.genos.zarr", p.name)
            if p.suffix != ".zarr"
            else p
            for c, p in paths.items()
        }
        return paths

    @classmethod
    def from_recs_genos(
        cls,
        records: Records,
        genotypes: Genotypes,
        paths: Optional[Union[str, Path, Dict[str, Path]]] = None,
        mem=int(1e9),
        overwrite: bool = False,
        chunk_shape: Optional[Tuple[int, int, int]] = None,
    ) -> Self:
        if not ZARR_TENSORSTORE_INSTALLED:
            raise ImportError(
                "Zarr and TensorStore must be installed to use chunked array caches like Zarr and N5."
            )

        n_samples = len(genotypes.samples)
        ploidy = genotypes.ploidy
        contigs = records.contigs
        contig_offsets = records.contig_offsets
        n_var_per_contig = {c: len(a) for c, a in records.v_starts.items()}

        if chunk_shape is None:
            chunk_shape = (10, ploidy, int(5e4))

        if paths is None:
            geno_paths = getattr(genotypes, "paths", None)
            if geno_paths is None:
                raise ValueError("No paths provided and no paths attribute found.")
            # suffix, possibly with .gz
            # examples: .pgen, .vcf, .vcf.gz, .bcf
            extension = re.compile(r"\.\w+(\.gz)?$")
            paths = {
                c: p.parent / extension.sub(".gvl.genos.zarr", p.name)
                for c, p in geno_paths.items()
            }
        elif isinstance(paths, (str, Path)):
            paths = {"_all": Path(paths)}

        if not overwrite and any(p.exists() for p in paths.values()):
            raise FileExistsError("Zarr file(s) already exist.")

        if "_all" in paths:
            one_source = True
        else:
            one_source = False

        first_path = next(iter(paths.values()))
        c_n_variants = np.array(list(n_var_per_contig.values()), dtype=np.int32)
        var_per_chunk = mem // (n_samples * ploidy)
        n_chunks = sum(np.ceil(c_n_variants / var_per_chunk).astype(int))
        if one_source:
            n_variants = sum(n_var_per_contig.values())
            ts_open_kwargs = {
                "spec": {
                    "driver": cls.driver,
                    "kvstore": {"driver": "file", "path": str(first_path)},
                },
                "dtype": np.int8,
                "shape": (n_samples, ploidy, n_variants),
                "create": True,
                "delete_existing": True,
                "chunk_layout": ts.ChunkLayout(  # pyright: ignore[reportPossiblyUnboundVariable,reportAttributeAccessIssue]
                    chunk_shape=chunk_shape
                ),
            }
            ts_handle = ts.open(  # pyright: ignore[reportPossiblyUnboundVariable,reportAttributeAccessIssue]
                **ts_open_kwargs
            ).result()
            arr = zarr.open_array(first_path)  # pyright: ignore[reportPossiblyUnboundVariable]
            arr.attrs["contigs"] = records.contigs
            arr.attrs["samples"] = genotypes.samples.tolist()

        with tqdm(total=n_chunks, unit="chunk") as pbar:
            for contig, c_n_vars in zip(contigs, c_n_variants):
                pbar.set_description(f"Reading contig {contig}")
                c_offset = contig_offsets[contig]

                if not one_source:
                    path = paths[contig]
                    ts_open_kwargs = {
                        "spec": {
                            "driver": cls.driver,
                            "kvstore": {"driver": "file", "path": str(path)},
                        },
                        "dtype": np.int8,
                        "shape": (n_samples, ploidy, c_n_vars),
                        "create": True,
                        "delete_existing": True,
                        "chunk_layout": ts.ChunkLayout(  # pyright: ignore[reportPossiblyUnboundVariable,reportAttributeAccessIssue]
                            chunk_shape=chunk_shape
                        ),
                    }
                    ts_handle = ts.open(  # pyright: ignore[reportPossiblyUnboundVariable,reportAttributeAccessIssue]
                        **ts_open_kwargs
                    ).result()
                    arr = zarr.open_array(path)  # pyright: ignore[reportPossiblyUnboundVariable]
                    arr.attrs["contigs"] = [contig]
                    arr.attrs["samples"] = genotypes.samples.tolist()

                # write all genotypes to cache in chunks of up to 1 GB of memory, using up
                # to (8 + 1) * chunk size = 9 GB working space
                # [c_offset, c_offset + c_n_vars)
                # for contig files, c_offset always = 0
                # for one file, c_offset = cumsum of previous offsets
                # c_n_vars is the number of variants in the contig
                idxs = np.arange(
                    c_offset, c_offset + c_n_vars + var_per_chunk, var_per_chunk
                )
                idxs[-1] = c_offset + c_n_vars
                for s_idx, e_idx in zip(idxs[:-1], idxs[1:]):
                    genos = genotypes.read(contig, np.array([s_idx]), np.array([e_idx]))
                    pbar.set_description(f"Writing contig {contig}")
                    ts_handle[  # pyright: ignore[reportPossiblyUnboundVariable]
                        :, :, s_idx:e_idx
                    ] = genos
                    pbar.update()

        return cls(paths)

    def _init_tstores(self):
        if "_all" in self.paths:
            one_source = True
        else:
            one_source = False

        first_path = next(iter(self.paths.values()))
        if one_source:
            tstore = self._open_tstore(next(iter(self.paths)))
            contigs = zarr.open_array(first_path).attrs["contigs"]  # pyright: ignore[reportPossiblyUnboundVariable]
            return {c: tstore for c in contigs}
        else:
            return {contig: self._open_tstore(contig) for contig in self.paths}

    def _open_tstore(self, contig: str):
        ts_open_kwargs = {
            "spec": {
                "driver": self.driver,
                "kvstore": {"driver": "file", "path": str(self.paths[contig])},
            },
            "open": True,
            "read": True,
        }

        return ts.open(  # pyright: ignore[reportPossiblyUnboundVariable,reportAttributeAccessIssue]
            **ts_open_kwargs
        ).result()

    def read(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
    ) -> NDArray[np.int8]:
        if self.tstores is None:
            self.tstores = self._init_tstores()

        start_idxs = np.atleast_1d(np.asarray(start_idxs, dtype=int))
        end_idxs = np.atleast_1d(np.asarray(end_idxs, dtype=int))

        # let tensorstore have the full query available to hopefully parallelize
        # reading variable length slices of genotypes
        sub_genos = [None] * len(start_idxs)
        tstore = self.tstores[contig]

        if sample_idx is not None:
            sample_idx = np.atleast_1d(np.asarray(sample_idx, dtype=int))
            tstore = tstore[sample_idx]
        if haplotype_idx is not None:
            haplotype_idx = np.atleast_1d(np.asarray(haplotype_idx, dtype=int))
            tstore = tstore[:, haplotype_idx]

        for i, (s, e) in enumerate(zip(start_idxs, end_idxs)):
            # no variants in query regions
            if s == e:
                continue
            sub_genos[i] = tstore[..., s:e]

        genotypes = ts.concat(  # pyright: ignore[reportPossiblyUnboundVariable,reportAttributeAccessIssue]
            sub_genos, axis=-1
        )[
            ts.d[0].translate_to[0]  # pyright: ignore[reportPossiblyUnboundVariable,reportAttributeAccessIssue]
        ]

        genotypes = cast(NDArray[np.int8], genotypes.read().result())

        return genotypes

    def vidx(
        self,
        contigs: ArrayLike,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idxs: ArrayLike,
        haplotype_idxs: ArrayLike,
    ) -> NDArray[np.int8]:
        if self.tstores is None:
            self.tstores = self._init_tstores()

        contigs = np.atleast_1d(np.asarray(contigs, dtype=np.str_))
        start_idxs = np.atleast_1d(np.asarray(start_idxs, dtype=np.intp))
        end_idxs = np.atleast_1d(np.asarray(end_idxs, dtype=np.intp))
        sample_idxs = np.atleast_1d(np.asarray(sample_idxs, dtype=np.intp))
        haplotype_idxs = np.atleast_1d(np.asarray(haplotype_idxs, dtype=np.intp))

        sub_genos = [None] * len(contigs)
        for i, (c, s, e, sp, h) in enumerate(
            zip(contigs, start_idxs, end_idxs, sample_idxs, haplotype_idxs)
        ):
            tstore = self.tstores[c]
            sub_genos[i] = tstore[sp, h, s:e]
        genotypes = ts.concat(  # pyright: ignore[reportPossiblyUnboundVariable,reportAttributeAccessIssue]
            sub_genos, axis=-1
        )[
            ts.d[0].translate_to[0]  # pyright: ignore[reportPossiblyUnboundVariable,reportAttributeAccessIssue]
        ]

        genotypes = cast(NDArray[np.int8], genotypes.read().result())

        return genotypes


class NumpyGenos(Genotypes, FromRecsGenos):
    chunked = False

    def __init__(
        self,
        arrays: Dict[str, NDArray[np.int8]],
        contig_offsets: Dict[str, int],
        samples: ArrayLike,
    ) -> None:
        self.contigs = list(arrays)
        self.arrays = arrays
        self.samples = np.array(samples, dtype=np.str_)
        self.ploidy = arrays[self.contigs[0]].shape[1]
        self.contig_offsets = contig_offsets

    def read(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
    ) -> NDArray[np.int8]:
        offset = self.contig_offsets[contig]
        start_idxs = np.atleast_1d(np.asarray(start_idxs, dtype=int)) - offset
        end_idxs = np.atleast_1d(np.asarray(end_idxs, dtype=int)) - offset

        if sample_idx is None:
            _sample_idx = slice(None)
            n_samples = len(self.samples)
        else:
            _sample_idx = np.atleast_1d(np.asarray(sample_idx, dtype=int))
            n_samples = len(_sample_idx)

        if haplotype_idx is None:
            _haplotype_idx = slice(None)
            ploidy = self.ploidy
        else:
            _haplotype_idx = np.atleast_1d(np.asarray(haplotype_idx, dtype=int))
            ploidy = len(_haplotype_idx)

        genos = np.empty(
            (n_samples, ploidy, (end_idxs - start_idxs).sum()),
            dtype=np.int8,
        )
        rel_starts = get_rel_starts(start_idxs, end_idxs)
        rel_ends = rel_starts + (end_idxs - start_idxs)
        for s, e, r_s, r_e in zip(start_idxs, end_idxs, rel_starts, rel_ends):
            if s == e:
                continue
            genos[..., r_s:r_e] = self.arrays[contig][_sample_idx, _haplotype_idx, s:e]

        return genos

    def __getitem__(
        self,
        contigs: ArrayLike,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idxs: ArrayLike,
        haplotype_idxs: ArrayLike,
    ) -> NDArray[np.int8]:
        contigs = np.atleast_1d(np.asarray(contigs, dtype=np.str_))
        start_idxs = np.atleast_1d(np.asarray(start_idxs, dtype=np.intp))
        end_idxs = np.atleast_1d(np.asarray(end_idxs, dtype=np.intp))
        sample_idxs = np.atleast_1d(np.asarray(sample_idxs, dtype=np.intp))
        haplotype_idxs = np.atleast_1d(np.asarray(haplotype_idxs, dtype=np.intp))

        genotypes = np.empty((end_idxs - start_idxs).sum(), dtype=np.int8)
        rel_starts = get_rel_starts(start_idxs, end_idxs)
        rel_ends = rel_starts + (end_idxs - start_idxs)
        for c, s, e, sp, h, r_s, r_e in zip(
            contigs,
            start_idxs,
            end_idxs,
            sample_idxs,
            haplotype_idxs,
            rel_starts,
            rel_ends,
        ):
            arr = self.arrays[c]
            offset = self.contig_offsets[c]
            genotypes[r_s:r_e] = arr[sp, h, s - offset : e - offset]

        return genotypes

    @classmethod
    def from_recs_genos(cls, records: Records, genotypes: Genotypes) -> Self:
        arrays = {}
        for contig in records.contigs:
            n_variants = len(records.v_starts[contig])
            arrays[contig] = genotypes.read(contig, 0, n_variants)
        return cls(arrays, records.contig_offsets, genotypes.samples)


class MemmapGenos(Genotypes, FromRecsGenos):
    chunked = False

    def __init__(self, memmaps: Dict[str, Path]) -> None:
        # TODO: need to store sample names as associated metadata
        raise NotImplementedError
        self.paths = memmaps
        self.memmaps = {
            c: np.memmap(p, dtype=np.int8, mode="r") for c, p in memmaps.items()
        }

    def read(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
    ) -> NDArray[np.int8]:
        raise NotImplementedError
        if sample_idx is None:
            _sample_idx = slice(None)
        else:
            _sample_idx = sample_idx
        if haplotype_idx is None:
            _haplotype_idx = slice(None)
        else:
            _haplotype_idx = haplotype_idx
        genos_ls: List[NDArray[np.int8]] = []
        for s_idx, e_idx in zip(start_idxs, end_idxs):
            # (s p v)
            genos_ls.append(
                self.memmaps[contig][_sample_idx, _haplotype_idx, s_idx:e_idx]
            )
        genos = np.concatenate(genos_ls, axis=-1)
        return genos

    @classmethod
    def from_recs_genos(
        cls,
        records: Records,
        genotypes: Genotypes,
        paths: Optional[Dict[str, Path]] = None,
        mem=int(1e9),
        overwrite: bool = False,
    ) -> Self:
        raise NotImplementedError
        n_samples = len(genotypes.samples)
        ploidy = genotypes.ploidy
        contig_offsets = records.contig_offsets
        n_var_per_contig = {c: len(a) for c, a in records.v_starts.items()}

        # TODO: get default paths if not provided
        memmaps: Dict[str, np.memmap] = {}
        for contig, path in paths.items():
            c_offset = contig_offsets[contig]
            c_n_vars = n_var_per_contig[contig]
            memmaps[contig] = np.memmap(
                path, dtype=np.int8, mode="w+", shape=(n_samples, ploidy, c_n_vars)
            )
            idxs = np.arange(c_offset, c_offset + c_n_vars + mem, mem)
            idxs[-1] = c_offset + c_n_vars
            for s_idx, e_idx in tqdm(
                zip(idxs[:-1], idxs[1:]),
                total=len(idxs) - 1,
                desc=f"Writing contig {contig}",
            ):
                memmaps[contig][..., s_idx:e_idx] = genotypes.read(
                    contig, np.array([s_idx]), np.array([e_idx])
                )[:]
                memmaps[contig].flush()
        return cls(paths)


class VCFGenos(Genotypes):
    chunked = False
    ploidy = 2

    def __init__(
        self, vcfs: Union[str, Path, Dict[str, Path]], contig_offsets: Dict[str, int]
    ) -> None:
        if not CYVCF2_INSTALLED:
            raise ImportError(
                "cyvcf2 must be installed to use VCF files for genotypes."
            )

        if isinstance(vcfs, (str, Path)):
            vcfs = {"_all": Path(vcfs)}

        self.paths = vcfs
        samples = None
        for p in self.paths.values():
            if not p.exists():
                raise FileNotFoundError(f"VCF file {p} not found.")
            if samples is None:
                samples = np.array(cyvcf2.VCF(str(p)).samples)  # pyright: ignore
        self.samples = samples  # type: ignore
        self.handles: Optional[Dict[str, cyvcf2.VCF]] = None  # pyright: ignore
        self.contig_offsets = contig_offsets

    def read(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
    ) -> NDArray[np.int8]:
        start_idxs = np.atleast_1d(np.asarray(start_idxs, dtype=int))
        end_idxs = np.atleast_1d(np.asarray(end_idxs, dtype=int))

        if self.handles is None and "_all" not in self.paths:
            self.handles = {
                c: cyvcf2.VCF(str(p), lazy=True)  # pyright: ignore
                for c, p in self.paths.items()
            }
        elif self.handles is None:
            handle = cyvcf2.VCF(str(next(iter(self.paths.values()))), lazy=True)  # pyright: ignore
            self.handles = {c: handle for c in handle.seqnames}

        if sample_idx is None:
            _sample_idx = slice(None)
        else:
            _sample_idx = np.atleast_1d(np.asarray(sample_idx, dtype=int))

        if haplotype_idx is None:
            _haplotype_idx = slice(None)
        else:
            _haplotype_idx = np.atleast_1d(np.asarray(haplotype_idx, dtype=int))

        # (s p v)
        genos = np.empty(
            (len(self.samples), self.ploidy, (end_idxs - start_idxs).sum()),
            dtype=np.int8,
        )
        geno_idx = 0
        offset = self.contig_offsets[contig]
        for i, v in enumerate(self.handles[contig](contig), start=offset):
            for s_idx, e_idx in zip(start_idxs, end_idxs):
                if i >= s_idx and i < e_idx:
                    # (s p) = (s p)
                    genos[..., geno_idx] = v.genotype.array()[:, :2]
                    geno_idx += 1
                if i >= e_idx:
                    break

        genos = genos[_sample_idx, _haplotype_idx]
        # cyvcf2 encoding: 0, 1, -1 => gvl/pgen encoding: 0, 1, -9
        genos[genos == -1] = -9

        return genos

    def __del__(self) -> None:
        if self.handles is not None:
            for handle in self.handles.values():
                handle.close()



from itertools import accumulate, chain, repeat
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

import dask.array as da
import numpy as np
import pandera as pa
import pandera.typing as pat
import polars as pl
import xarray as xr
from numpy.typing import NDArray

from .types import Reader

T = TypeVar("T")


def construct_virtual_data(
    *readers: Reader, n_regions: int, fixed_length: int
) -> xr.Dataset:
    arrays = {}
    for reader in readers:
        dims = ["region"] + list(reader.sizes) + ["length"]
        shape = [n_regions] + [size for size in reader.sizes.values()] + [fixed_length]
        arrays[reader.name] = xr.DataArray(
            da.empty(  # pyright: ignore[reportPrivateImportUsage]
                shape, dtype=reader.dtype
            ),
            dims=dims,
            coords=reader.coords,
            name=reader.name,
        )
    virtual_data = xr.Dataset(arrays)
    return virtual_data


def process_bed(bed: Union[str, Path, pl.DataFrame], fixed_length: int):
    if isinstance(bed, (str, Path)):
        bed = read_bedlike(bed)

    if "strand" in bed and bed["strand"].dtype == pl.Utf8:
        bed = bed.with_columns(
            pl.col("strand").replace({"-": -1, "+": 1}, return_dtype=pl.Int8)
        )
    else:
        bed = bed.with_columns(strand=pl.lit(1, dtype=pl.Int8))

    if "region_idx" not in bed:
        bed = bed.with_row_count("region_idx")

    bed = bed.sort("chrom", "chromStart")

    return _set_fixed_length_around_center(bed, fixed_length)


def _set_fixed_length_around_center(bed: pl.DataFrame, length: int):
    if "peak" in bed:
        center = pl.col("chromStart") + pl.col("peak")
    else:
        center = (pl.col("chromStart") + pl.col("chromEnd")) / 2
    bed = bed.with_columns(
        chromStart=(center - length / 2).round(0).cast(pl.Int64),
        chromEnd=(center + length / 2).round(0).cast(pl.Int64),
    )
    return bed


def random_chain(
    *iterables: Iterable[T], seed: Optional[int] = None
) -> Generator[T, None, None]:
    """Chain iterables, randomly sampling from each until they are all exhausted."""
    rng = np.random.default_rng(seed)
    iterators = {i: iter(it) for i, it in enumerate(iterables)}
    while iterators:
        i = rng.choice(list(iterators.keys()))
        try:
            yield next(iterators[i])
        except StopIteration:
            del iterators[i]


def read_bedlike(path: Union[str, Path]) -> pl.DataFrame:
    """Reads a bed-like (BED3+) file as a pandas DataFrame. The file type is inferred
    from the file extension.

    Parameters
    ----------
    path : Union[str, Path]

    Returns
    -------
    polars.DataFrame
    """
    path = Path(path)
    if ".bed" in path.suffixes:
        return _read_bed(path)
    elif ".narrowPeak" in path.suffixes:
        return _read_narrowpeak(path)
    elif ".broadPeak" in path.suffixes:
        return _read_broadpeak(path)
    else:
        try:
            return _read_bed_table(path)
        except ValueError:
            raise ValueError(
                f"""Unrecognized file extension: {''.join(path.suffixes)}. Expected one 
                of .bed, .narrowPeak, .broadPeak, or a table file (e.g. .csv, .tsv)"""
            )


class BEDSchema(pa.DataFrameModel):
    chrom: pat.Series[str]
    chromStart: pat.Series[int]
    chromEnd: pat.Series[int]
    name: Optional[pat.Series[str]] = pa.Field(nullable=True)
    score: Optional[pat.Series[float]] = pa.Field(nullable=True)
    strand: Optional[pat.Series[str]] = pa.Field(isin=["+", "-", "."], nullable=True)
    thickStart: Optional[pat.Series[int]] = pa.Field(nullable=True)
    thickEnd: Optional[pat.Series[int]] = pa.Field(nullable=True)
    itemRgb: Optional[pat.Series[str]] = pa.Field(nullable=True)
    blockCount: Optional[pat.Series[pa.UInt]] = pa.Field(nullable=True)
    blockSizes: Optional[pat.Series[str]] = pa.Field(nullable=True)
    blockStarts: Optional[pat.Series[str]] = pa.Field(nullable=True)

    class Config:
        coerce = True


def _read_bed(bed_path: Union[str, Path]):
    with open(bed_path) as f:
        skip_rows = 0
        while (line := f.readline()).startswith(("track", "browser")):
            skip_rows += 1
    n_cols = line.count("\t") + 1
    bed_cols = [
        "chrom",
        "chromStart",
        "chromEnd",
        "name",
        "score",
        "strand",
        "thickStart",
        "thickEnd",
        "itemRgb",
        "blockCount",
        "blockSizes",
        "blockStarts",
    ]
    bed = pl.read_csv(
        bed_path,
        separator="\t",
        has_header=False,
        skip_rows=skip_rows,
        new_columns=bed_cols[:n_cols],
        dtypes={"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8},
        null_values=".",
    ).to_pandas()
    bed = BEDSchema.to_schema()(bed)
    return pl.from_pandas(bed)


def _read_bed_table(table: Union[str, Path], **table_reader_kwargs):
    table = Path(table)
    suffixes = set(table.suffixes)
    reader_kwargs: Dict[str, Any] = {}
    reader_kwargs.update(table_reader_kwargs)
    if ".csv" in suffixes:
        reader_kwargs["separator"] = ","
        reader_kwargs["dtypes"] = {"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8}
        reader = pl.scan_csv
    elif {".txt", ".tsv"} & suffixes:
        reader_kwargs["separator"] = "\t"
        reader_kwargs["dtypes"] = {"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8}
        reader = pl.scan_csv
    elif {".fth", ".feather", ".ipc", ".arrow"} & suffixes:
        reader = pl.scan_ipc  # type : ignore[assignment]
    else:
        raise ValueError(f"Table has unrecognized file extension: {table.name}")
    bed = reader(table, **reader_kwargs).collect().to_pandas()
    bed = BEDSchema.to_schema()(bed)
    return pl.from_pandas(bed)


class NarrowPeakSchema(pa.DataFrameModel):
    chrom: pat.Series[str]
    chromStart: pat.Series[int]
    chromEnd: pat.Series[int]
    name: pat.Series[str] = pa.Field(nullable=True)
    score: pat.Series[float] = pa.Field(nullable=True)
    strand: pat.Series[str] = pa.Field(isin=["+", "-", "."], nullable=True)
    signalValue: pat.Series[float] = pa.Field(nullable=True)
    pValue: pat.Series[float] = pa.Field(nullable=True)
    qValue: pat.Series[float] = pa.Field(nullable=True)
    peak: pat.Series[int] = pa.Field(nullable=True)

    class Config:
        coerce = True


def _read_narrowpeak(narrowpeak_path: Union[str, Path]):
    with open(narrowpeak_path) as f:
        skip_rows = 0
        while f.readline().startswith(("track", "browser")):
            skip_rows += 1
    narrowpeaks = pl.read_csv(
        narrowpeak_path,
        separator="\t",
        has_header=False,
        skip_rows=skip_rows,
        new_columns=[
            "chrom",
            "chromStart",
            "chromEnd",
            "name",
            "score",
            "strand",
            "signalValue",
            "pValue",
            "qValue",
            "peak",
        ],
        dtypes={"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8},
        null_values=".",
    ).to_pandas()
    narrowpeaks = NarrowPeakSchema.to_schema()(narrowpeaks)
    return pl.from_pandas(narrowpeaks)


class BroadPeakSchema(pa.DataFrameModel):
    chrom: pat.Series[str]
    chromStart: pat.Series[int]
    chromEnd: pat.Series[int]
    name: pat.Series[str] = pa.Field(nullable=True)
    score: pat.Series[float] = pa.Field(nullable=True)
    strand: pat.Series[str] = pa.Field(isin=["+", "-", "."], nullable=True)
    signalValue: pat.Series[float] = pa.Field(nullable=True)
    pValue: pat.Series[float] = pa.Field(nullable=True)
    qValue: pat.Series[float] = pa.Field(nullable=True)

    class Config:
        coerce = True


def _read_broadpeak(broadpeak_path: Union[str, Path]):
    with open(broadpeak_path) as f:
        skip_rows = 0
        while f.readline().startswith(("track", "browser")):
            skip_rows += 1
    broadpeaks = pl.read_csv(
        broadpeak_path,
        separator="\t",
        has_header=False,
        skip_rows=skip_rows,
        new_columns=[
            "chrom",
            "chromStart",
            "chromEnd",
            "name",
            "score",
            "strand",
            "signalValue",
            "pValue",
            "qValue",
        ],
        dtypes={"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8},
        null_values=".",
    ).to_pandas()
    broadpeaks = BroadPeakSchema.to_schema()(broadpeaks)
    return pl.from_pandas(broadpeaks)


def _cartesian_product(arrays: Sequence[NDArray]) -> NDArray:
    """Get the cartesian product of multiple arrays such that each entry corresponds to
    a unique combination of the input arrays' values.
    """
    # https://stackoverflow.com/a/49445693
    la = len(arrays)
    shape = *map(len, arrays), la
    dtype = np.result_type(*arrays)
    arr = np.empty(shape, dtype=dtype)
    arrs = (*accumulate(chain((arr,), repeat(0, la - 1)), np.ndarray.__getitem__),)
    idx = slice(None), *repeat(None, la - 1)
    for i in range(la - 1, 0, -1):
        arrs[i][..., i] = arrays[i][idx[: la - i]]
        arrs[i - 1][1:] = arrs[i]
    arr[..., 0] = arrays[0][idx]
    return arr.reshape(-1, la)


def get_rel_starts(
    starts: NDArray[np.int64], ends: NDArray[np.int64]
) -> NDArray[np.int64]:
    rel_starts: NDArray[np.int64] = np.concatenate(
        [[0], (ends - starts).cumsum()[:-1]], dtype=np.int64
    )
    return rel_starts


DTYPE = TypeVar("DTYPE", bound=np.generic)


def splice_subarrays(
    arr: NDArray[DTYPE], starts: NDArray[np.int64], ends: NDArray[np.int64]
) -> NDArray[DTYPE]:
    """Splice subarrays from a larger array.

    Parameters
    ----------
    arr : NDArray
        Array to splice from.
    starts : NDArray[np.int64]
        Start coordinates, 0-based.
    ends : NDArray[np.int64]
        End coordinates, 0-based exclusive.

    Returns
    -------
    out : NDArray
        Spliced array.
    """
    start = starts.min()
    rel_starts = get_rel_starts(starts, ends)
    total_length = (ends - starts).sum()
    shape = arr.shape[:-1] + (total_length,)
    out = np.empty(shape, dtype=arr.dtype)
    for rel_start, s, e in zip(rel_starts, starts - start, ends - start):
        length = e - s
        out[..., rel_start : rel_start + length] = arr[..., s:e]
    return out


def splice_and_rc_subarrays(
    arr: NDArray[DTYPE],
    starts: NDArray[np.int64],
    ends: NDArray[np.int64],
    strands: NDArray[np.int8],
    rc_fn: Callable[[NDArray[DTYPE]], NDArray[DTYPE]],
) -> NDArray[DTYPE]:
    """Splice subarrays from a larger array and reverse-complement them.

    Parameters
    ----------
    arr : NDArray
        Array to splice from.
    starts : NDArray[np.int64]
        Start coordinates, 0-based.
    ends : NDArray[np.int64]
        End coordinates, 0-based exclusive.
    strands : NDArray[np.int8]
        Strand of each query region. 1 for forward, -1 for reverse. If None, defaults
        to forward strand.
    rc_fn : Callable[[NDArray], NDArray]
        Function to reverse-complement the subarrays.

    Returns
    -------
    out : NDArray
        Spliced and reverse-complemented array.
    """
    start = starts.min()
    rel_starts = get_rel_starts(starts, ends)
    total_length = (ends - starts).sum()
    shape = arr.shape[:-1] + (total_length,)
    out = np.empty(shape, dtype=arr.dtype)
    for rel_start, s, e, strand in zip(
        rel_starts, starts - start, ends - start, strands
    ):
        length = e - s
        if strand == 1:
            subarr = arr[..., s:e]
        else:
            subarr = rc_fn(arr[..., s:e])
        out[..., rel_start : rel_start + length] = subarr
    return out


def splice_and_rev_subarrays(
    arr: NDArray[DTYPE],
    starts: NDArray[np.int64],
    ends: NDArray[np.int64],
    strands: NDArray[np.int8],
) -> NDArray[DTYPE]:
    """
    Splices and reverses subarrays of a given array based on the start and end indices
    and strand orientation.

    Parameters
    ----------
    arr : NDArray
        Array to splice from.
    starts : NDArray[np.int64]
        Start coordinates, 0-based.
    ends : NDArray[np.int64]
        End coordinates, 0-based exclusive.
    strands : NDArray[np.int8]
        Strand of each query region. 1 for forward, -1 for reverse. If None, defaults
        to forward strand.

    Returns
    -------
    out : NDArray
        Spliced and reversed array.
    """
    start = starts.min()
    rel_starts = get_rel_starts(starts, ends)
    total_length = (ends - starts).sum()
    shape = arr.shape[:-1] + (total_length,)
    out = np.empty(shape, dtype=arr.dtype)
    for rel_start, s, e, strand in zip(
        rel_starts, starts - start, ends - start, strands
    ):
        length = e - s
        out[..., rel_start : rel_start + length] = arr[..., s:e:strand]
    return out


def normalize_contig_name(contig: str, contigs: Iterable[str]) -> Optional[str]:
    """Normalize the contig name to adhere to the convention of the underlying file.
    i.e. remove or add "chr" to the contig name.

    Parameters
    ----------
    contig : str

    Returns
    -------
    str
        Normalized contig name.
    """
    for c in contigs:
        # exact match, remove chr, add chr
        if contig == c or contig[3:] == c or f"chr{contig}" == c:
            return c
    return None



import re
from pathlib import Path
from textwrap import dedent
from typing import Dict, Iterable, List, Optional, Tuple, Union, overload

import numba as nb
import numpy as np
import polars as pl
from attrs import define
from loguru import logger
from numpy.typing import ArrayLike, NDArray
from tqdm.auto import tqdm
from typing_extensions import Self

try:
    import cyvcf2

    CYVCF2_INSTALLED = True
except ImportError:
    CYVCF2_INSTALLED = False


@define
class VLenAlleles:
    """Variable length alleles.

    Create VLenAlleles from a polars Series of strings:
    >>> alleles = VLenAlleles.from_polars(pl.Series(["A", "AC", "G"]))

    Create VLenAlleles from offsets and alleles:
    >>> offsets = np.array([0, 1, 3, 4], np.uint32)
    >>> alleles = np.frombuffer(b"AACG", "|S1")
    >>> alleles = VLenAlleles(offsets, alleles)

    Get a single allele:
    >>> alleles[0]
    b'A'

    Get a slice of alleles:
    >>> alleles[1:]
    VLenAlleles(offsets=array([0, 2, 3]), alleles=array([b'AC', b'G'], dtype='|S1'))
    """

    offsets: NDArray[np.uint32]
    alleles: NDArray[np.bytes_]

    @overload
    def __getitem__(self, idx: int) -> NDArray[np.bytes_]:
        ...

    @overload
    def __getitem__(self, idx: slice) -> "VLenAlleles":
        ...

    def __getitem__(self, idx: Union[int, slice, np.integer]):
        if isinstance(idx, (int, np.integer)):
            return self.get_idx(idx)
        elif isinstance(idx, slice):
            return self.get_slice(idx)

    def get_idx(self, idx: Union[int, np.integer]):
        if idx >= len(self) or idx < -len(self):
            raise IndexError("Index out of range.")
        if idx < 0:
            idx = len(self) + idx
        return self.alleles[self.offsets[idx] : self.offsets[idx + 1]]

    def get_slice(self, slc: slice):
        start: Optional[int]
        stop: Optional[int]
        start, stop = slc.start, slc.stop
        if start is None:
            start = 0
        elif start < 0:
            start += len(self)

        # handle empty result
        if start >= len(self) or (stop is not None and stop <= start):
            return VLenAlleles(np.empty(0, np.uint32), np.empty(0, "|S1"))

        if stop is None:
            stop = len(self)
        elif stop < 0:
            stop += len(self)
        stop += 1
        new_offsets = self.offsets[start:stop].copy()
        _start, _stop = new_offsets[0], new_offsets[-1]
        new_alleles = self.alleles[_start:_stop]
        new_offsets -= self.offsets[start]
        return VLenAlleles(new_offsets, new_alleles)

    def __len__(self):
        """Number of alleles."""
        return len(self.offsets) - 1

    @property
    def nbytes(self):
        return self.offsets.nbytes + self.alleles.nbytes

    @classmethod
    def from_polars(cls, alleles: pl.Series):
        offsets = np.r_[np.uint32(0), alleles.str.len_bytes().cum_sum().to_numpy()]
        flat_alleles = np.frombuffer(
            alleles.str.concat("").to_numpy()[0].encode(), "S1"
        )
        return cls(offsets, flat_alleles)

    @staticmethod
    def concat(*vlen_alleles: "VLenAlleles"):
        if len(vlen_alleles) == 0:
            return VLenAlleles(np.array([0], np.uint32), np.array([], "|S1"))
        elif len(vlen_alleles) == 1:
            return vlen_alleles[0]

        offset_ends = np.array([v.offsets[-1] for v in vlen_alleles[:-1]]).cumsum(
            dtype=np.uint32
        )
        offsets = np.concatenate(
            [
                vlen_alleles[0].offsets,
                *(
                    v.offsets + last_offset_end
                    for v, last_offset_end in zip(vlen_alleles[1:], offset_ends)
                ),
            ]
        )
        alleles = np.concatenate([v.alleles for v in vlen_alleles])
        return VLenAlleles(offsets, alleles)


@define
class RecordInfo:
    positions: NDArray[np.int32]
    size_diffs: NDArray[np.int32]
    refs: VLenAlleles
    alts: VLenAlleles
    start_idxs: NDArray[np.int32]
    end_idxs: NDArray[np.int32]
    offsets: NDArray[np.uint32]


@define
class Records:
    contigs: List[str]
    contig_offsets: Dict[str, int]

    ## sorted by starts ##
    v_starts: Dict[str, NDArray[np.int32]]
    # difference in length between ref and alt, sorted by start
    v_diffs: Dict[str, NDArray[np.int32]]
    # no multi-allelics
    ref: Dict[str, VLenAlleles]
    alt: Dict[str, VLenAlleles]
    ######################

    ## sorted by ends ##
    v_ends: Dict[str, NDArray[np.int32]]
    e2s_idx: Dict[str, NDArray[np.int32]]
    v_diffs_sorted_by_ends: Dict[str, NDArray[np.int32]]
    max_del_q: Dict[str, NDArray[np.intp]]
    ####################

    @property
    def n_variants(self) -> int:
        return sum(len(v) for v in self.v_starts.values())

    @classmethod
    def from_vcf(cls, vcf: Union[str, Path, Dict[str, Path]]) -> Self:
        if not CYVCF2_INSTALLED:
            raise ImportError(
                "cyvcf2 is not installed. Please install it with `pip install cyvcf2`"
            )

        if isinstance(vcf, (str, Path)):
            vcf = {"_all": Path(vcf)}

        if "_all" in vcf:
            multi_contig_source = True
        else:
            multi_contig_source = False

        vcf_suffix = re.compile(r"\.[vb]cf(\.gz)?$")
        arrow_paths = {
            c: p.parent / vcf_suffix.sub(".gvl.arrow", p.name) for c, p in vcf.items()
        }

        if cls.gvl_arrow_exists(vcf, arrow_paths):
            return cls.from_gvl_arrow(arrow_paths)

        if multi_contig_source:
            start_df = cls.read_vcf(vcf["_all"])
            start_dfs = start_df.partition_by("#CHROM", as_dict=True)
            del start_df
        else:
            start_dfs: Dict[str, pl.DataFrame] = {}
            for contig, path in vcf.items():
                start_dfs[contig] = cls.read_vcf(path)

        start_dfs, end_dfs = cls.process_start_df(start_dfs)

        if multi_contig_source:
            path = vcf["_all"]
            pl.concat(start_dfs.values()).write_ipc(
                path.parent / vcf_suffix.sub(".gvl.arrow", path.name)
            )
            pl.concat(end_dfs.values()).write_ipc(
                path.parent / vcf_suffix.sub(".gvl.ends.arrow", path.name)
            )
        else:
            for s_df, e_df, path in zip(
                start_dfs.values(), end_dfs.values(), vcf.values()
            ):
                s_df.write_ipc(path.parent / vcf_suffix.sub(".gvl.arrow", path.name))
                e_df.write_ipc(
                    path.parent / vcf_suffix.sub(".gvl.ends.arrow", path.name)
                )

        return cls.from_var_df(start_dfs, end_dfs)

    @staticmethod
    def read_vcf(vcf_path: Path):
        if not CYVCF2_INSTALLED:
            raise ImportError(
                "cyvcf2 is not installed. Please install it with `pip install cyvcf2`"
            )

        vcf = cyvcf2.VCF(
            str(vcf_path)
        )  # pyright: ignore[reportPossiblyUnboundVariable]
        n_variants = vcf.num_records
        chroms = [None] * n_variants
        positions = np.empty(n_variants, dtype=np.int32)
        refs = [None] * n_variants
        alts = [None] * n_variants
        non_snp_non_indel = False
        with tqdm(total=n_variants, desc=f"Reading {vcf_path.name}") as pbar:
            for i, v in enumerate(vcf):
                if not v.is_snp and not v.is_indel:
                    non_snp_non_indel = True
                    continue
                chroms[i] = v.CHROM
                positions[i] = v.POS
                refs[i] = v.REF
                alt = v.ALT
                # TODO: punt multi-allelics. also punt missing ALT?
                if len(alt) != 1:
                    raise RuntimeError(
                        f"""VCF file {vcf_path} contains multi-allelic or overlappings
                        variants which are not yet supported by GenVarLoader. Normalize 
                        the VCF with `bcftools norm -f <reference.fa>
                        -a --atom-overlaps . -m - <file.vcf>`"""
                    )
                alts[i] = alt[0]
                pbar.update()
        if non_snp_non_indel:
            logger.warning(
                f"""VCF file {vcf_path} contains non-SNP and non-INDEL variants. 
                These variants will be ignored."""
            )
        return pl.DataFrame(
            {
                "#CHROM": chroms,
                "POS": positions,
                "REF": refs,
                "ALT": alts,
            }
        )

    @classmethod
    def from_pvar(cls, pvar: Union[str, Path, Dict[str, Path]]) -> Self:
        if isinstance(pvar, (str, Path)):
            pvar = {"_all": Path(pvar)}

        if "_all" in pvar:
            multi_contig_source = True
        else:
            multi_contig_source = False

        arrow_paths = {c: p.with_suffix(".gvl.arrow") for c, p in pvar.items()}

        if cls.gvl_arrow_exists(pvar, arrow_paths):
            return cls.from_gvl_arrow(arrow_paths)

        if multi_contig_source:
            start_df = cls.read_pvar(pvar["_all"])
            start_dfs = start_df.partition_by("#CHROM", as_dict=True)
        else:
            start_dfs: Dict[str, pl.DataFrame] = {}
            for contig, path in pvar.items():
                start_dfs[contig] = cls.read_pvar(path)

        start_dfs, end_dfs = cls.process_start_df(start_dfs)

        if multi_contig_source:
            path = pvar["_all"]
            pl.concat(start_dfs.values()).write_ipc(path.with_suffix(".gvl.arrow"))
            pl.concat(end_dfs.values()).write_ipc(path.with_suffix(".gvl.ends.arrow"))
        else:
            for s_df, e_df, path in zip(
                start_dfs.values(), end_dfs.values(), pvar.values()
            ):
                s_df.write_ipc(path.with_suffix(".gvl.arrow"))
                e_df.write_ipc(path.with_suffix(".gvl.ends.arrow"))

        return cls.from_var_df(start_dfs, end_dfs)

    @staticmethod
    def read_pvar(pvar_path: Path):
        with open(pvar_path, "r") as f:
            skip_rows = 0
            while f.readline().startswith("##"):
                skip_rows += 1

        logger.info("Reading .pvar file...")
        pvar = pl.read_csv(
            pvar_path,
            separator="\t",
            skip_rows=skip_rows,
            columns=["#CHROM", "POS", "REF", "ALT"],
            dtypes={"#CHROM": pl.Utf8, "POS": pl.Int32},
        )
        logger.info("Finished reading .pvar file.")
        if (pvar["ALT"].str.contains(",")).any():
            raise RuntimeError(
                f"""PGEN file {pvar_path} contains multi-allelic variants which are 
                not yet supported by GenVarLoader. Split your multi-allelic 
                variants with `bcftools norm -f <reference.fa> -a
                --atom-overlaps . -m - <file.vcf>` then remake the PGEN file
                with the `--vcf-half-call r` option."""
            )
        return pvar

    @staticmethod
    def gvl_arrow_exists(
        sources: Union[Path, Dict[str, Path]], arrow: Union[Path, Dict[str, Path]]
    ) -> bool:
        # TODO: check if files were created after the last breaking change to the gvl.arrow format
        # check if the files exist
        if isinstance(arrow, Path):
            assert isinstance(sources, Path)
            if not arrow.exists():
                return False
            if not arrow.stat().st_mtime >= sources.stat().st_mtime:
                return False
        else:
            assert isinstance(sources, dict)
            for contig, p in arrow.items():
                if not p.exists():
                    return False
                if not p.stat().st_mtime > sources[contig].stat().st_mtime:
                    return False

        return True

    @classmethod
    def from_gvl_arrow(cls, arrow: Union[str, Path, Dict[str, Path]]) -> Self:
        if isinstance(arrow, (str, Path)):
            arrow = {"_all": Path(arrow)}

        if "_all" in arrow:
            multi_contig_source = True
        else:
            multi_contig_source = False

        arrow_paths: Dict[str, Path] = {}
        for c, p in arrow.items():
            if p.suffix == ".gvl":
                arrow_paths[c] = p.with_suffix(".gvl.arrow")
            elif p.suffixes[-2:] == [".gvl", ".arrow"]:
                arrow_paths[c] = p
            else:
                raise ValueError(
                    f"Arrow file {p} does not have the .gvl suffix. Please provide the correct path."
                )

        if multi_contig_source:
            path = arrow_paths["_all"]
            start_dfs = pl.read_ipc(path).partition_by("#CHROM", as_dict=True)
            end_dfs = pl.read_ipc(
                path.parent / path.name.replace(".gvl.arrow", ".gvl.ends.arrow")
            ).partition_by("#CHROM", as_dict=True)
        else:
            start_dfs: Dict[str, pl.DataFrame] = {}
            end_dfs: Dict[str, pl.DataFrame] = {}
            for contig, path in arrow_paths.items():
                start_dfs[contig] = pl.read_ipc(path)
                end_dfs[contig] = pl.read_ipc(
                    path.parent / path.name.replace(".gvl.arrow", ".gvl.ends.arrow")
                )

        return cls.from_var_df(start_dfs, end_dfs)

    @staticmethod
    def process_start_df(start_dfs: Dict[str, pl.DataFrame]):
        """_summary_

        Parameters
        ----------
        start_dfs : Dict[str, pl.DataFrame]
            Each dataframe should have columns: POS, REF, ALT

        Returns
        -------
        start_dfs : Dict[str, pl.DataFrame]
        end_dfs : Dict[str, pl.DataFrame]
        """
        end_dfs: Dict[str, pl.DataFrame] = {}
        for contig, df in start_dfs.items():
            df = df.with_columns(
                POS=pl.col("POS") - 1,  #! change to 0-indexed
                ILEN=(
                    pl.col("ALT").str.len_bytes().cast(pl.Int32)
                    - pl.col("REF").str.len_bytes().cast(pl.Int32)
                ),
            )
            start_dfs[contig] = df

            ends = (
                df.select(
                    "#CHROM",
                    "ILEN",
                    END=pl.col("POS")
                    - pl.col("ILEN").clip(upper_bound=0),  #! end-inclusive
                )
                .with_row_count("VAR_IDX")
                .select(
                    pl.all().sort_by("END"),
                    # make E2S_IDX relative to each contig
                    pl.int_range(0, pl.count(), dtype=pl.UInt32)
                    .sort_by("END")
                    .reverse()
                    .rolling_min(df.height, min_periods=1)
                    .reverse()
                    .alias("E2S_IDX"),
                )
                .select("#CHROM", "END", "ILEN", "VAR_IDX", "E2S_IDX")
                .sort("END")
                .join(df.select("POS").with_row_count("VAR_IDX"), on="VAR_IDX")
            )

            _starts = ends["POS"].to_numpy()
            _ends = ends["END"].to_numpy()
            was_ends = np.empty(len(_ends) + 1, dtype=_ends.dtype)
            was_ends[0] = 0
            #! convert to end-exclusive, + 1
            was_ends[1:] = _ends + 1
            md_q = np.searchsorted(was_ends, _starts, side="right") - 1
            ends = ends.with_columns(MD_Q=md_q).select(
                "#CHROM", "END", "MD_Q", "ILEN", "E2S_IDX"
            )

            end_dfs[contig] = ends

        return start_dfs, end_dfs

    @classmethod
    def from_var_df(
        cls, start_dfs: Dict[str, pl.DataFrame], end_dfs: Dict[str, pl.DataFrame]
    ) -> Self:
        ## sorted by starts ##
        v_starts: Dict[str, NDArray[np.int32]] = {}
        # difference in length between ref and alt, sorted by start
        v_diffs: Dict[str, NDArray[np.int32]] = {}
        # no multi-allelics
        ref: Dict[str, VLenAlleles] = {}
        alt: Dict[str, VLenAlleles] = {}
        ######################

        ## sorted by ends ##
        v_ends: Dict[str, NDArray[np.int32]] = {}
        v_diffs_sorted_by_ends: Dict[str, NDArray[np.int32]] = {}
        e2s_idx: Dict[str, NDArray[np.int32]] = {}
        max_del_q: Dict[str, NDArray[np.int32]] = {}
        ####################

        contig_offset = 0
        contig_offsets: Dict[str, int] = {}
        for contig, s_df in start_dfs.items():
            e_df = end_dfs[contig]
            contig_offsets[contig] = contig_offset
            contig_offset += s_df.height
            v_starts[contig] = s_df["POS"].to_numpy()
            v_diffs[contig] = s_df["ILEN"].to_numpy()
            v_ends[contig] = e_df["END"].to_numpy()
            v_diffs_sorted_by_ends[contig] = e_df["ILEN"].to_numpy()
            e2s_idx[contig] = np.empty(e_df.height + 1, dtype=np.uint32)
            e2s_idx[contig][:-1] = e_df["E2S_IDX"].to_numpy()
            e2s_idx[contig][-1] = e_df.height
            max_del_q[contig] = e_df["MD_Q"].to_numpy()

            # no multi-allelics
            ref[contig] = VLenAlleles.from_polars(s_df["REF"])
            alt[contig] = VLenAlleles.from_polars(s_df["ALT"])

        return cls(
            contigs=list(start_dfs.keys()),
            contig_offsets=contig_offsets,
            v_starts=v_starts,
            v_diffs=v_diffs,
            ref=ref,
            alt=alt,
            v_ends=v_ends,
            v_diffs_sorted_by_ends=v_diffs_sorted_by_ends,
            e2s_idx=e2s_idx,
            max_del_q=max_del_q,
        )

    def vars_in_range(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
    ) -> Optional[RecordInfo]:
        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = np.atleast_1d(np.asarray(ends, dtype=int))

        _s_idxs = np.searchsorted(self.v_ends[contig], starts)
        # make idxs absolute
        s_idxs = self.e2s_idx[contig][_s_idxs] + self.contig_offsets[contig]
        e_idxs = (
            np.searchsorted(self.v_starts[contig], ends) + self.contig_offsets[contig]
        )

        if s_idxs.min() == e_idxs.max():
            return None

        n_var_per_region = e_idxs - s_idxs
        offsets = np.empty(len(n_var_per_region) + 1, dtype=np.uint32)
        offsets[0] = 0
        offsets[1:] = np.cumsum(n_var_per_region)

        rel_s_idxs = s_idxs - self.contig_offsets[contig]
        rel_e_idxs = e_idxs - self.contig_offsets[contig]

        positions = np.concatenate(
            [self.v_starts[contig][s:e] for s, e in zip(rel_s_idxs, rel_e_idxs)]
        )
        size_diffs = np.concatenate(
            [self.v_diffs[contig][s:e] for s, e in zip(rel_s_idxs, rel_e_idxs)]
        )
        ref = VLenAlleles.concat(
            *(self.ref[contig][s:e] for s, e in zip(rel_s_idxs, rel_e_idxs))
        )
        alt = VLenAlleles.concat(
            *(self.alt[contig][s:e] for s, e in zip(rel_s_idxs, rel_e_idxs))
        )

        return RecordInfo(
            positions=positions,
            size_diffs=size_diffs,
            refs=ref,
            alts=alt,
            start_idxs=s_idxs,
            end_idxs=e_idxs,
            offsets=offsets,
        )

    def vars_in_range_for_haplotype_construction(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
    ) -> Tuple[Optional[RecordInfo], NDArray[np.int32]]:
        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = np.atleast_1d(np.asarray(ends, dtype=int))

        _s_idxs = np.searchsorted(self.v_ends[contig], starts)

        max_ends, _e_idxs = get_max_ends_and_idxs(
            self.v_ends[contig],
            self.v_diffs_sorted_by_ends[contig],
            self.max_del_q[contig],
            _s_idxs,
            ends,
        )

        s_idxs = self.e2s_idx[contig][_s_idxs]
        e_idxs = self.e2s_idx[contig][_e_idxs]

        # make idxs absolute
        s_idxs += self.contig_offsets[contig]
        e_idxs += self.contig_offsets[contig]

        if s_idxs.min() == e_idxs.max():
            return None, ends.astype(np.int32)

        np.concatenate(
            [np.arange(s, e, dtype=np.uint32) for s, e in zip(s_idxs, e_idxs)]
        )
        n_var_per_region = e_idxs - s_idxs
        offsets = np.empty(len(n_var_per_region) + 1, dtype=np.uint32)
        offsets[0] = 0
        np.cumsum(n_var_per_region, out=offsets[1:])

        rel_s_idxs = s_idxs - self.contig_offsets[contig]
        rel_e_idxs = e_idxs - self.contig_offsets[contig]

        positions = np.concatenate(
            [self.v_starts[contig][s:e] for s, e in zip(rel_s_idxs, rel_e_idxs)]
        )
        size_diffs = np.concatenate(
            [self.v_diffs[contig][s:e] for s, e in zip(rel_s_idxs, rel_e_idxs)]
        )
        ref = VLenAlleles.concat(
            *(self.ref[contig][s:e] for s, e in zip(rel_s_idxs, rel_e_idxs))
        )
        alt = VLenAlleles.concat(
            *(self.alt[contig][s:e] for s, e in zip(rel_s_idxs, rel_e_idxs))
        )

        v_info = RecordInfo(
            positions=positions,
            size_diffs=size_diffs,
            refs=ref,
            alts=alt,
            start_idxs=s_idxs,
            end_idxs=e_idxs,
            offsets=offsets,
        )

        return v_info, max_ends

    def vidx_vars_in_range(
        self,
        contigs: ArrayLike,
        starts: ArrayLike,
        length: int,
    ) -> Optional[RecordInfo]:
        contigs = np.atleast_1d(np.asarray(contigs, dtype=str))
        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = starts + length
        n_queries = len(contigs)

        positions = np.empty(n_queries, np.int32)
        size_diffs = np.empty(n_queries, np.int32)
        refs: List[VLenAlleles] = [
            None
        ] * n_queries  # pyright: ignore[reportAssignmentType]
        alts: List[VLenAlleles] = [
            None
        ] * n_queries  # pyright: ignore[reportAssignmentType]
        start_idxs = np.empty(n_queries, np.int32)
        end_idxs = np.empty(n_queries, np.int32)

        unique_contigs = np.unique(contigs)
        for contig in unique_contigs:
            idx = np.nonzero(contigs == contig)[0]

            _s_idxs = np.searchsorted(self.v_ends[contig], starts)
            rel_s_idxs = self.e2s_idx[contig][_s_idxs]
            rel_e_idxs = np.searchsorted(self.v_starts[contig], ends)

            for i, r_s, r_e in zip(idx, rel_s_idxs, rel_e_idxs):
                positions[i] = self.v_starts[contig][r_s:r_e]
                size_diffs[i] = self.v_diffs[contig][r_s:r_e]
                refs[i] = self.ref[contig][r_s:r_e]
                alts[i] = self.alt[contig][r_s:r_e]

            # make idxs absolute
            s_idxs = rel_s_idxs + self.contig_offsets[contig]
            e_idxs = rel_e_idxs + self.contig_offsets[contig]
            start_idxs[idx] = s_idxs
            end_idxs[idx] = e_idxs

        n_var_per_region = end_idxs - start_idxs
        offsets = np.empty(len(n_var_per_region) + 1, dtype=np.uint32)
        offsets[0] = 0
        offsets[1:] = np.cumsum(n_var_per_region)

        return RecordInfo(
            positions=positions,
            size_diffs=size_diffs,
            refs=VLenAlleles.concat(*refs),
            alts=VLenAlleles.concat(*alts),
            start_idxs=start_idxs,
            end_idxs=end_idxs,
            offsets=offsets,
        )

    def vidx_vars_in_range_for_haplotype_construction(
        self,
        contigs: ArrayLike,
        starts: ArrayLike,
        length: int,
    ) -> Tuple[Optional[RecordInfo], NDArray[np.int32]]:
        contigs = np.atleast_1d(np.asarray(contigs, dtype=str))
        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = starts + length
        n_queries = len(contigs)

        positions = np.empty(n_queries, np.int32)
        size_diffs = np.empty(n_queries, np.int32)
        refs: List[VLenAlleles] = [
            None
        ] * n_queries  # pyright: ignore[reportAssignmentType]
        alts: List[VLenAlleles] = [
            None
        ] * n_queries  # pyright: ignore[reportAssignmentType]
        start_idxs = np.empty(n_queries, np.int32)
        end_idxs = np.empty(n_queries, np.int32)

        unique_contigs = np.unique(contigs)
        max_ends = np.empty(n_queries, np.int32)
        for contig in unique_contigs:
            idx = np.nonzero(contigs == contig)[0]

            _s_idxs = np.searchsorted(self.v_ends[contig], starts)
            _max_ends, _e_idxs = get_max_ends_and_idxs(
                self.v_ends[contig],
                self.v_diffs_sorted_by_ends[contig],
                self.max_del_q[contig],
                _s_idxs,
                ends,
            )
            rel_s_idxs = self.e2s_idx[contig][_s_idxs]
            rel_e_idxs = self.e2s_idx[contig][_e_idxs]

            for i, r_s, r_e in zip(idx, rel_s_idxs, rel_e_idxs):
                positions[i] = self.v_starts[contig][r_s:r_e]
                size_diffs[i] = self.v_diffs[contig][r_s:r_e]
                refs[i] = self.ref[contig][r_s:r_e]
                alts[i] = self.alt[contig][r_s:r_e]

            # make idxs absolute
            s_idxs = rel_s_idxs + self.contig_offsets[contig]
            e_idxs = rel_e_idxs + self.contig_offsets[contig]

            start_idxs[idx] = s_idxs
            end_idxs[idx] = e_idxs
            max_ends[idx] = _max_ends

        n_var_per_region = end_idxs - start_idxs
        offsets = np.empty(len(n_var_per_region) + 1, dtype=np.uint32)
        offsets[0] = 0
        offsets[1:] = np.cumsum(n_var_per_region)

        return (
            RecordInfo(
                positions=positions,
                size_diffs=size_diffs,
                refs=VLenAlleles.concat(*refs),
                alts=VLenAlleles.concat(*alts),
                start_idxs=start_idxs,
                end_idxs=end_idxs,
                offsets=offsets,
            ),
            max_ends,
        )

    def normalize_contig_name(
        self, contig: str, contigs: Iterable[str]
    ) -> Optional[str]:
        """Normalize the contig name to adhere to the convention of the underlying file.
        i.e. remove or add "chr" to the contig name.

        Parameters
        ----------
        contig : str

        Returns
        -------
        str
            Normalized contig name.
        """
        for c in contigs:
            # exact match, remove chr, add chr
            if contig == c or contig[3:] == c or f"chr{contig}" == c:
                return c
        return None

    def __repr__(self):
        return dedent(
            f"""
            {self.__class__.__name__}
            contigs: {self.contigs}
            n_variants: {self.n_variants})
            """
        ).strip()


@nb.njit(nogil=True, cache=True)
def get_max_ends_and_idxs(
    v_ends: NDArray[np.int32],
    v_diffs: NDArray[np.int32],
    nearest_nonoverlapping: NDArray[np.intp],
    start_idxs: NDArray[np.intp],
    query_ends: NDArray[np.int64],
) -> Tuple[NDArray[np.int32], NDArray[np.intp]]:
    """Time: O(q * v) where q is number of query regions and v is the number of variants
    (mostly limited to those in the query regions thanks to early exit criteria).
    Importantly, this algorithm does not scale (directly) with the number of samples. However, more samples
    from sequencing experiemnts will increase the number of unique variants sub-linearly.
    """
    max_ends: NDArray[np.int32] = np.empty(len(start_idxs), dtype=np.int32)
    end_idxs: NDArray[np.intp] = np.empty(len(start_idxs), dtype=np.intp)
    for r in nb.prange(len(start_idxs)):
        s = start_idxs[r]
        if s == len(v_ends):  # no variants in this region
            max_ends[r] = query_ends[r]
            end_idxs[r] = s
            continue

        w = -v_diffs[s:]  # flip sign so deletions have positive weight

        # to adjust q from [0, j) to [i, j)
        # (q[i:] - i).clip(0)
        q = nearest_nonoverlapping[s:]
        max_end, end_idx = weighted_activity_selection(
            v_ends[s:], w, q, s, query_ends[r]
        )
        max_ends[r] = max_end
        end_idxs[r] = s + end_idx
    return max_ends, end_idxs


@nb.njit(nogil=True, cache=True)
def weighted_activity_selection(
    v_ends: NDArray[np.int32],
    w: NDArray[np.int32],
    q: NDArray[np.intp],
    start_index: int,
    query_end: int,
) -> Tuple[int, int]:
    """Implementation of the [weighted activity selection problem](https://en.wikipedia.org/wiki/Activity_selection_problem)
    to compute the maximum length of deletions that can occur for each region. This is
    used to adjust the end coordinates for reference sequence queries and include all
    variants for that are needed for haplotype construction.

    Parameters
    ----------
    v_ends : NDArray[np.int64]
        Shape: (variants). End coordinates of variants, 0-based inclusive.
    w : NDArray[np.int64]
        Shape: (variants). Weights of activities (i.e. deletion lengths).
    q : NDArray[np.intp]
        Shape: (variants). Nearest variant i such that i < j and variants are non-overlapping, q[j] = i.
        Note this should not yet be adjusted for the start_index.
    start_index : int
        Index of the first variant overlapping the query region.
    query_end : int
        End of query region.

    Returns
    -------
    max_ends : NDArray[np.int32]
        Shape: (regions). Maximum end coordinate for each query region.
    end_idxs : NDArray[np.intp]
        Shape: (regions). Index of the variant with the maximum end coordinate for each
        query region.

    Notes
    -----
    For the weighted activity selection problem, each deletion corresponds
    to an activity with weight equal to the length of the deletion. The goal is to
    compute the maximum total weight of deletions for each query region.

    Psuedocode from (Princeton slides)[https://www.cs.princeton.edu/~wayne/cs423/lectures/dynamic-programming-4up.pdf]:
    Given starts :math: `s_1, ..., s_n`, ends :math: `e_1, ..., e_n`, and weights
    :math: `w_1, ..., w_n`.
    Note that ends are sorted, :math: `e_1 <= ... <= e_n`.
    Let :math: `q_j` = largest index :math: `i < j` such that activity :math: `i` is
    compatible with :math: `j`.
    Let opt(j) = value of solution to the problem consisting of activities 1 to j
    Then,
        opt(0) = 0
    and
        opt(j) = max(w_j + opt(q_j), opt(j - 1))
    """
    n_vars = len(w)
    max_del: NDArray[np.int32] = np.empty(n_vars + 1, dtype=np.int32)
    max_del[0] = 0
    for j in range(1, n_vars + 1):
        _q = max(q[j - 1] - start_index, 0)
        max_del[j] = max(max_del[_q] + w[j - 1], max_del[j - 1])
        v_del_end = v_ends[j - 1] - max_del[j] + 1  # + 1, v_ends is end-inclusive
        # if:
        # this variant more than satisfies query length
        # last variant doesn't span v_del_end
        if v_del_end > query_end and j > 1 and v_ends[j - 2] <= v_del_end:
            # then add the max deletion length up to but not including this variant
            # to the query end, and return the index of this variant for slicing
            return query_end + max_del[j - 1], j - 1
    return query_end + max_del[-1], n_vars



from itertools import zip_longest
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np
from numpy.typing import ArrayLike, NDArray
from tqdm.auto import tqdm

from .types import Reader
from .util import normalize_contig_name

try:
    import tensorstore as ts
    import zarr

    ZARR_TENSORSTORE_INSTALLED = True
except ImportError:
    ZARR_TENSORSTORE_INSTALLED = False


"""
Implementation note regarding using tensorstore with num_workers > 0
https://github.com/google/tensorstore/issues/61
TL;DR TensorStore is not Send.
"""


class ZarrTracks(Reader):
    chunked = True

    def __init__(self, name: str, path: Union[str, Path]):
        if not ZARR_TENSORSTORE_INSTALLED:
            raise ImportError(
                "ZarrTracks support requires zarr and tensorstore to be installed."
            )

        self.name = name
        self.path = Path(path)
        z = zarr.open_group(self.path, mode="r")  # pyright: ignore
        self.contigs = cast(Dict[str, int], z.attrs["contigs"])
        self.sizes = cast(Dict[str, int], z.attrs["sizes"])
        self.coords = {d: np.asarray(z.attrs[d]) for d in self.sizes}
        self.samples = cast(Optional[List[str]], z.attrs.get("sample", None))
        self.ploidy = cast(Optional[int], z.attrs.get("ploid", None))
        self.dtype = z[next(iter(self.contigs))].dtype
        # each is (s? p? l)
        self.tstores: Optional[Dict[str, Any]] = None

    def rev_strand_fn(self, x):
        return x[..., ::-1]

    @classmethod
    def from_reader(
        cls,
        reader: Reader,
        path: Union[str, Path],
        chunk_shape: Optional[Tuple[int, ...]] = None,
    ):
        if isinstance(path, str):
            path = Path(path)
        z = zarr.open_group(path)  # pyright: ignore[reportPossiblyUnboundVariable]
        z.attrs["contigs"] = reader.contigs
        z.attrs["sizes"] = reader.sizes
        if "sample" in reader.coords:
            z.attrs["sample"] = reader.coords["sample"].tolist()
        if "ploid" in reader.sizes:
            z.attrs["ploid"] = reader.sizes["ploidy"]
        with tqdm(total=len(reader.contigs)) as pbar:
            for contig, e in reader.contigs.items():
                pbar.set_description(f"Reading {contig}")
                data = reader.read(contig, 0, e)
                pbar.set_description(f"Writing {contig}")
                if chunk_shape is None:
                    chunk_layout = ts.ChunkLayout(  # pyright: ignore
                        chunk_shape=(20,) * len(reader.sizes) + (int(5e4),)
                    )
                else:
                    chunk_layout = ts.ChunkLayout(  # pyright: ignore
                        chunk_shape=chunk_shape
                    )
                tstore = ts.open(  # pyright: ignore
                    {
                        "driver": "zarr",
                        "kvstore": {"driver": "file", "path": str(path / contig)},
                        "metadata": {
                            "compressor": {
                                "id": "blosc",
                                "shuffle": -1,
                                "clevel": 5,
                                "cname": "lz4",
                            }
                        },
                    },
                    read=True,
                    write=True,
                    create=True,
                    delete_existing=True,
                    dtype=data.dtype,
                    chunk_layout=chunk_layout,
                    shape=data.shape,
                ).result()
                tstore.write(data).result()
                pbar.update()
        zarr.consolidate_metadata(str(path))  # pyright: ignore

        return cls(reader.name, path)

    def _tstore(self, contig: str):
        tstore = ts.open(  # pyright: ignore
            {
                "driver": "zarr",
                "kvstore": {"driver": "file", "path": str(self.path / contig)},
            },
            read=True,
        ).result()
        return tstore

    def read(
        self, contig: str, starts: ArrayLike, ends: ArrayLike, **kwargs
    ) -> NDArray:
        if self.tstores is None:
            self.tstores = {contig: self._tstore(contig) for contig in self.contigs}

        contig = normalize_contig_name(
            contig, self.contigs
        )  # pyright: ignore[reportAssignmentType]
        if contig is None:
            raise ValueError(f"Contig {contig} not found")

        tstore = self.tstores[contig]
        samples = cast(Optional[List[str]], kwargs.get("sample", self.samples))
        if samples is not None:
            if self.samples is None:
                samples = None
            elif missing := set(samples).difference(self.samples):
                raise ValueError(f"Samples {missing} were not found")
            else:
                key_idx, query_idx = np.intersect1d(
                    self.samples, samples, return_indices=True
                )[1:]
                sample_idx = key_idx[query_idx]
                # (s p? l)
                tstore = tstore[sample_idx]

        ploidy = cast(Optional[ArrayLike], kwargs.get("ploid", None))
        if ploidy is not None:
            if self.ploidy is None:
                ploidy = None
            else:
                haplotype_idx = np.asarray(ploidy, dtype=int)
                if (haplotype_idx >= self.ploidy).any():
                    raise ValueError("Ploidies requested exceed maximum ploidy")
                # (s? p l)
                if self.samples is None:
                    # (p l)
                    tstore = tstore[haplotype_idx]
                else:
                    # (s p l)
                    tstore = tstore[:, haplotype_idx]

        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = np.atleast_1d(np.asarray(ends, dtype=int))

        sub_values = [None] * len(starts)
        for i, (s, e) in enumerate(zip(starts, ends)):
            # no variants in query regions
            if s == e:
                continue
            # (s? p? l)
            sub_values[i] = tstore[..., s:e]

        values = ts.concat(sub_values, axis=-1)[  # pyright: ignore  # pyright: ignore
            ts.d[0].translate_to[0]  # pyright: ignore
        ]

        values = cast(NDArray, values.read().result())

        return values

    def vidx(
        self,
        contigs: ArrayLike,
        starts: ArrayLike,
        length: int,
        samples: Optional[ArrayLike] = None,
        ploidy: Optional[ArrayLike] = None,
        **kwargs,
    ) -> NDArray:
        if self.tstores is None:
            self.tstores = {contig: self._tstore(contig) for contig in self.contigs}

        contigs = np.array(
            [
                normalize_contig_name(c, self.contigs)
                for c in np.atleast_1d(np.asarray(contigs))
            ]
        )
        if (contigs == None).any():  # noqa: E711
            raise ValueError("Some contigs not found")

        if samples is not None:
            if self.samples is None:
                raise ValueError("No sample information available")

            unique_samples, inverse = np.unique(samples, return_inverse=True)
            if missing := set(unique_samples).difference(self.samples):
                raise ValueError(f"Samples {missing} were not found")

            key_idx, query_idx = np.intersect1d(
                self.samples, unique_samples, return_indices=True, assume_unique=True
            )[1:]
            sample_idx = key_idx[query_idx[inverse]]
        else:
            if self.samples is None:
                sample_idx = [None]
            else:
                sample_idx = [slice(None)]

        if ploidy is not None:
            if self.ploidy is None:
                raise ValueError("No ploidy information available")

            haplotype_idx = np.asarray(ploidy, dtype=int)
            if (haplotype_idx >= self.ploidy).any():
                raise ValueError("Ploidies requested exceed maximum ploidy")
        else:
            if self.ploidy is None:
                haplotype_idx = [None]
            else:
                haplotype_idx = [slice(None)]

        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = starts + length

        sub_values = [None] * len(starts)
        for i, (c, s, e, sp, h) in enumerate(
            zip_longest(contigs, starts, ends, sample_idx, haplotype_idx)
        ):
            # no variants in query regions
            if s == e:
                continue
            tstore = self.tstores[c]
            # (s? p? l)
            if sp is None and h is None:
                # (l)
                sub_values[i] = tstore[s:e]
            elif sp is None:
                # (p l)
                sub_values[i] = tstore[h, s:e]
            elif h is None:
                # (s p l)
                sub_values[i] = tstore[sp, s:e]
            else:
                sub_values[i] = tstore[sp, h, s:e]

        values = ts.concat(sub_values, axis=-1)[  # pyright: ignore
            ts.d[0].translate_to[0]  # pyright: ignore
        ]

        values = cast(NDArray, values.read().result())

        return values
